/***************************************************************************
 *                                                                         *
 *   copyright (C) 2003, 2004 by Michael Buesch                            *
 *   email: mbuesch@freenet.de                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License version 2        *
 *   as published by the Free Software Foundation.                         *
 *                                                                         *
 ***************************************************************************/

#ifndef PWMKEYCARD_H
#define PWMKEYCARD_H

#include "cardif.h"

#include <qwidget.h>
#include <qtimer.h>
#include <string>

using std::string;

class WaitWnd;
class SpinForSignal;

/** PwM key-card (chipcard) interface class
  * An instance of this class deletes itself, when
  * its work has been done. So you don't have to worry
  * about deleting the object.
  */
class PwMKeyCard : public QObject
{
	Q_OBJECT
public:
	PwMKeyCard(QWidget *_parent = 0);
	~PwMKeyCard();

	/** generate a new key-card (all data will be lost!) */
	void genNewCard();
	/** write data to card */
	void writeNewCard(uint32_t cardId, const string &key);
	/** erase the currently inserted card (all data will be lost!) */
	void eraseCard();
	/** get the key stored on the current card */
	void getKey();
	/** display the key in a message-box */
	void displayKey();
	/** make backup image */
	void makeBackupImage();
	/** replay backup image */
	void replayBackupImage();

	/** the user may use this spinner to spin for a signal
	  * of this class
	  */
	SpinForSignal* getSpinner()
			{ return spinner; }

signals:
	/** key data requested by getKey() is available */
	void keyAvailable(uint32_t cardId, const string &key);

protected:
	/** tries to open a card */
	bool open();
	/** display opencard wait wnd */
	void openCardWaitWnd();
	/** remove opencard wait wnd */
	void removeOpenCardWaitWnd();
	/** display read-write card wait wnd */
	void rwCardWaitWnd();
	/** remove read-wrote card wait wnd */
	void removeRwCardWaitWnd();
	/** write backup-image file */
	void writeBackupImageFile(const QString &filename, const string &key,
				  uint32_t cardId);
	/** read data from backup-image file */
	bool readBackupImageFile(const QString &filename, string *key,
				 uint32_t *cardId);

protected slots:
	/** open() status slot */
	void open_status(cardif_status stat);
	/** close() status slot */
	void close_status(cardif_status);
	/** genNewCard() status slot */
	void genNewCard_status(cardif_status stat);
	/** writeNewCard() status slot */
	void writeNewCard_status(cardif_status stat);
	/** getKey() status slot */
	void getKey_status(cardif_status stat, uint32_t cardId,
			   const string &);
	/** eraseCard() status slot */
	void eraseCard_status(cardif_status stat);
	/** decrement the time count for open wait window */
	void decOpenWaitWndTime();
	/** loopback slot for displayKey() */
	void displayKey_slot(uint32_t cardId, const string &key);
	/** loopback slot for makeBackupImage() */
	void makeBackupImage_slot(uint32_t cardId, const string &key);
	/** cancel operation. This may be invoked by the user */
	void cancelOp();

protected:
	/** parent widget */
	QWidget *parent;
	/** lowlevel card interface */
	CardIf *cardIf;
	/** current task the user requested */
	cardif_task currentTask;
	/** open wait window */
	WaitWnd *openWaitWnd;
	/** read-write wait window */
	WaitWnd *rwWaitWnd;
	/** timer for open wait window */
	QTimer *openWaitWnd_timer;
	/** current timeout count for open wait window */
	int curOpenWaitWnd_timeout;
	/** are we accessing the card-terminal */
	static bool ctAccess;
	/** spin for a signal (if needed) */
	SpinForSignal *spinner;
	/** tempoary */
	string tmp_str;
	/** tempoary */
	uint32_t tmp_u32;
};

#endif
