/*************************************************************************
 *                                                                       *
 *   copyright (C) 2003, 2004 by Michael Buesch                          *
 *   email: mbuesch@freenet.de                                           *
 *                                                                       *
 *  This library is free software; you can redistribute it and/or        *
 *  modify it under the terms of the GNU Lesser General Public           *
 *  License version 2.1 as published by the Free Software Foundation.    *
 *                                                                       *
 *  This library is distributed in the hope that it will be useful,      *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU    *
 *  Lesser General Public License for more details.                      *
 *                                                                       *
 *  You should have received a copy of the GNU Lesser General Public     *
 *  License along with this library; if not, write to the Free Software  *
 *  Foundation, Inc., 59 Temple Place,                                   *
 *  Suite 330, Boston, MA 02111-1307  USA                                *
 *                                                                       *
 *************************************************************************/

/* BIG FAT NOTE!!!
 * This interface is not frozen and EXPERIMENTAL, yet!
 * That means, there will be changes in the future to the interface
 * (API and ABI), that break things. The rivision numbers will not
 * be increased for changes, while the interface is not frozen.
 */

#ifndef __LIB_PWMANAGERDCOP_H
#define __LIB_PWMANAGERDCOP_H

#include <qstring.h>
#include <qobject.h>

#define PWMANAGER_DCOP_REV_MAJOR	(static_cast<unsigned int>(0x00))
#define PWMANAGER_DCOP_REV_MINOR	(static_cast<unsigned int>(0x00))

class PwManagerDCOPif_stub;


/**
  * PwManager  --  Password manager
  *
  * public DCOP interface class
  *
  * @author Michael Buesch <mbuesch@freenet.de>
  * @short PwManager DCOP interface
  */
class PwManagerDCOP : public QObject
{
	Q_OBJECT
public:
	/** Constructor
	  * @param parent Parent QObject
	  * @param name Name-string
	  */
	PwManagerDCOP(QObject *parent = 0, const char *name = 0);
	/** Destructor */
	virtual ~PwManagerDCOP();

	/** struct entry_struct defines a logical
	  * password-entry.
	  */
	struct entry_struct
	{
		/** entry description */
		QString desc;
		/** user-name */
		QString name;
		/** the password */
		QString pw;
		/** some comment */
		QString comment;
		/** an URL string */
		QString url;
		/** launcher. Can be executed as a system() command */
		QString launcher;

		void clear()
		{
			desc = "";
			name = "";
			pw = "";
			comment = "";
			url = "";
			launcher = "";
		}
	};

	/** Version numbers.
	  * @return Returns the revision numbers
	  *	    (QPair(major, minor)) of the interface.
	  */
	QPair<unsigned int, unsigned int> version();
	/** Check if running.
	  * Please call this function first, to ensure PwManager
	  * is running.
	  * @return Returns true if PwManager is running.
	  */
	bool isRunning();
	/** Open-document list.
	  * @return Returns a QStringList of all currently opened
	  *	    documents (files).
	  */
	QStringList getDocumentList();
	/** Connect to an open document.
	  * Call this to autenticate yourself at PwManager.
	  * This opens a window which asks for which document to
	  * connect to.
	  * PwManager asks for the master-password on behalf
	  * of this function call. If you didn't call this function
	  * successfully, you don't have any access to the document data.
	  * @return Returns 0 on success, -1 if not connected, and >=1 on
	  *         other errors.
	  */
	int connectToDoc();
	/** Disconnect the current document-connection. */
	void disconnectDoc();
	/** Close the current document and disconnect
	  * the current document-connection.
	  */
	void closeConnectedDoc();
	/** Open a new document with the data in the given filename.
	  * PwManager asks for the master-password (or keycard) and
	  * displays message-boxes for errors, that occur.
	  * @param filename If an empty string is given for the filename,
	  *                 PwManager asks for which file to open.
	  * @return Returns 0 on success, -1 if not connected, and >=1 on
	  *         other errors.
	  */
	int openDocument(const QString &filename);
	/** PwManager opens a window which requests the user to change
	  * the password of the DCOP-connected document.
	  */
	void changeMpw();
	/** Get a category-list.
	  * @return Returns a QStringList of all available categories
	  *         in the connected document.
	  */
	QStringList getCategoryList();
	/** Rename a category.
	  * @param category The old category name, as given in getCategoryList()
	  * @param newName The new category name.
	  * @return Returns 0 on success, -1 if not connected, and >=1 on
	  *         other errors.
	  */
	int renameCategory(const QString &category, const QString &newName);
	/** Delete / Remove a category and all of its entries.
	  * @param category The category name to delete.
	  * @return Returns 0 on success, -1 if not connected, and >=1 on
	  *         other errors.
	  */
	int deleteCategory(const QString &category);
	/** Get an entry-list.
	  * @param category The category of which you want to
	  *                 generate an entry-list.
	  * @return Returns a QStringList of all entries in the given
	  *         category.
	  */
	QStringList getEntryList(const QString &category);
	/** Add an entry to a category.
	  * @param category The category to which you want to add the entry.
	  * @param entry The password-entry-data you want to add to category.
	  * @return Returns 0 on success, -1 if not connected, and >=1 on
	  *         other errors.
	  */
	int addEntry(const QString &category, const entry_struct &entry);
	/** Open add-entry-window.
	  * PwManager opens the add-entry-window and requests the user
	  * to add a new entry.
	  * @param pw The password initially displayed in the add-entry-window.
	  *           You may select an empty string here.
	  *           The user can still edit the password in the add-window.
	  */
	void addEntry_wnd(const QString &pw);
	/** Get the contents of an entry.
	  * @param category The category to which the entry belongs,
	  *                 that you want to get.
	  * @param index The index number of the QStringList, returned
	  *              by getEntryList(), that you want to get.
	  * @return Returns a struct entry_struct data struct of the
	  *         given entry.
	  */
	entry_struct getEntry(const QString &category, int index);
	/** Edit the data of an existing password entry.
	  * @param category The name of the category in which the entry
	  *                 is, that you want to edit.
	  * @param index The index number of the QStringList, returned
	  *              by getEntryList(), that you want to edit.
	  * @param newCategory The new Category which to move the entry
	  *                    to. If the category doesn't exist, it's
	  *                    created.
	  * @param newEntryData The new (edited) data for the entry.
	  * @return Returns 0 on success, -1 if not connected, and >=1 on
	  *         other errors.
	  */
	int editEntry(const QString &category, int index,
		      const QString &newCategory,
		      const entry_struct &newEntryData);
	/** Open edit-entry-window.
	  * PwManager opens the edit-entry window for the given entry.
	  * @param category The name of the category in which the entry
	  *                 is, that you want to edit.
	  * @param index The index number of the QStringList, returned
	  *              by getEntryList(), that you want to edit.
	  */
      	void editEntry_wnd(const QString &category, int index);
	/** Delete an entry.
	  * @param category The name of the category in which the entry
	  *                 is, that you want to delete.
	  * @param index The index number of the QStringList, returned
	  *              by getEntryList(), that you want to delete.
	  */
	int deleteEntry(const QString &category, int index);

protected:
	/** @internal
	  * DCOP stub object
	  */
	PwManagerDCOPif_stub *dcopStub;
};

#endif // __LIB_PWMANAGERDCOP_H
