/***************************************************************************
 *                                                                         *
 *   copyright (C) 2004 by Michael Buesch                                  *
 *   email: mbuesch@freenet.de                                             *
 *                                                                         *
 *   KWallet DCOP Interface is                                             *
 *    Copyright (c) 2002-2003 George Staikos <staikos@kde.org>             *
 *   KWallet is part of the KDE libraries                                  *
 *   Some code in this class is derived from the original KWallet          *
 *   code. This code is  copyright by George Staikos.                      *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License version 2        *
 *   as published by the Free Software Foundation.                         *
 *                                                                         *
 ***************************************************************************/

#include "kwalletemu.h"
#include "pwmexception.h"
#include "pwmdoc.h"
#include "randomizer.h"
#include "pwm.h"

#include <dcopclient.h>
#include <dcopref.h>
#include <kapplication.h>
#include <kconfig.h>
#include <kdatastream.h>

#include <qcstring.h>

#include <stdint.h>

using std::vector;


void KWalletEmu::ConnectionList::add(const QCString &client, PwMDoc *doc, int handle)
{
#ifdef PWM_DEBUG
	vector<connectionElement>::iterator
			i = connList.begin(),
			end = connList.end();
	while (i != end) {
		if (i->handle == handle) {
			BUG();
			return;
		}
		++i;
	}
#endif // PWM_DEBUG
	connectionElement newItem;
	newItem.clientName = client;
	newItem.document = doc;
	newItem.handle = handle;
	connList.push_back(newItem);
}

void KWalletEmu::ConnectionList::del(int handle)
{
	vector<connectionElement>::iterator
			i = connList.begin(),
			end = connList.end();
	while (i != end) {
		if (i->handle == handle) {
			kwalletEmu->sig_applicationDisconnected(i->document->getTitle(),
								i->clientName);
			connList.erase(i);
			return;
		}
		++i;
	}
	printDebug("KWalletEmu::ConnectionList::del(int handle):  element not found.");
}

void KWalletEmu::ConnectionList::del(const QString &wallet)
{
	vector<connectionElement>::iterator
			i = connList.begin(),
			end = connList.end();
	while (i != end) {
		if (i->document->getTitle() == wallet) {
			kwalletEmu->sig_applicationDisconnected(i->document->getTitle(),
								i->clientName);
			connList.erase(i);
		}
		++i;
	}
}

void KWalletEmu::ConnectionList::delClient(const QCString &client)
{
	vector<connectionElement>::iterator
			i = connList.begin(),
			end = connList.end();
	while (i != end) {
		if (i->clientName == client) {
			kwalletEmu->sig_applicationDisconnected(i->document->getTitle(),
								i->clientName);
			connList.erase(i);
		}
		++i;
	}
}

bool KWalletEmu::ConnectionList::get(int handle, connectionElement *ret)
{
	vector<connectionElement>::iterator
			i = connList.begin(),
			end = connList.end();
	while (i != end) {
		if (i->handle == handle) {
			if (ret)
				*ret = *i;
			return true;
		}
		++i;
	}
	return false;
}

bool KWalletEmu::ConnectionList::find(const QCString &client, const QString &wallet,
				      connectionElement *ret)
{
	vector<connectionElement>::iterator
			i = connList.begin(),
			end = connList.end();
	while (i != end) {
		if (i->clientName == client) {
			if (wallet == i->document->getTitle()) {
				if (ret)
					*ret = *i;
				return true;
			}
		}
		++i;
	}
	return false;
}

bool KWalletEmu::ConnectionList::find(const QString wallet,
				      connectionElement *ret)
{
	vector<connectionElement>::iterator
			i = connList.begin(),
			end = connList.end();
	while (i != end) {
		if (i->document->getTitle() == wallet) {
			if (ret)
				*ret = *i;
			return true;
		}
		++i;
	}
	return false;
}


KWalletEmu::KWalletEmu(const QCString &objId, QObject *parent,
		       PwM *_mainClass, const char *name)
 : QObject(parent, name)
 , DCOPObject(objId)
{
	if (!dcopClient) {
		throw PwMException(EX_GENERIC,
				   "dcop not available");
		return;
	}
	connect(dcopClient, SIGNAL(applicationRemoved(const QCString &)),
		this, SLOT(dcopAppDisconn(const QCString &)));
	mainClass = _mainClass;
	connList = new ConnectionList(this);
	setEnabled(true);

	if (!kdedUnloadKWallet()) {
		throw PwMException(EX_LOAD_MODULE,
				   "unload kwallet failed");
		return;
	}
	if (!kdedLoadEmu()) {
		throw PwMException(EX_LOAD_MODULE,
				   "load emu failed");
		return;
	}
	sig_walletListDirty();
}

KWalletEmu::~KWalletEmu()
{
	int ret;
	sig_allWalletsClosed();
	setEnabled(false);
	if (!kdedUnloadEmu()) {
		printWarn("unloading KDED KWallet emulation "
			  "module failed!");
	}
	ret = kdedLoadKWallet();
	if (ret < 0) {
		printWarn("(re)loading KDED KWallet "
			  "module failed!");
	}
	if (unlikely(connList->getList()->size())) {
		printWarn("There are KWallet connections left "
			  "after shutting down the emulation!");
	}
	delete connList;
}

void KWalletEmu::connectDocSignals(PwMDoc *doc)
{
	connect(doc, SIGNAL(dataChanged(PwMDoc *)),
		this, SLOT(docChanged(PwMDoc *)));
	connect(doc, SIGNAL(docClosed(PwMDoc *)),
		this, SLOT(docClosed(PwMDoc *)));
	connect(doc, SIGNAL(docOpened(PwMDoc *)),
		this, SLOT(docOpened(PwMDoc *)));
	connect(doc, SIGNAL(docCreated(PwMDoc *)),
		this, SLOT(docCreated(PwMDoc *)));
	sig_walletCreated(doc->getTitle());
	sig_walletListDirty();
}

QCString KWalletEmu::DCOPPeerName()
{
	DCOPClient *dc = callingDcopClient();
	if (!dc)
		return "";
	return dc->senderId();
}

int KWalletEmu::kdedLoadKWallet()
{
	KConfig kwalletConf("kwalletrc");
	kwalletConf.setGroup("Wallet");
	bool loadKWallet = kwalletConf.readBoolEntry("Enabled", true);
	if (!loadKWallet) {
		printDebug("KWallet module not "
			   "loaded because it's disabled");
		return 1;
	}

	bool ok;
	QCString replyType;
	QByteArray replyData;
	QByteArray param;
	QDataStream paramDs(param, IO_WriteOnly);
	paramDs << QCString("kwalletd");
	ok = dcopClient->call("kded", "kded",
			      "loadModule(QCString)",
			      param,
			      replyType,
			      replyData,
			      true);
	if (!ok) {
		printDebug("kded::loadModule(kwalletd) "
			   "dcop call failed");
		return -1;
	}
	if (replyType != "bool") {
		printWarn("kded::loadModule() return "
			  "type != bool");
		return -1;
	}
	bool boolRet;
	QDataStream replyStream(replyData, IO_ReadOnly);
	replyStream >> boolRet;
	if (!boolRet) {
		printDebug("loading kwalletd failed.");
		return -1;
	}
	return 0;
}

bool KWalletEmu::kdedLoadEmu()
{
	bool ok;
	QCString replyType;
	QByteArray replyData;
	QByteArray param;
	QDataStream paramDs(param, IO_WriteOnly);
	paramDs << QCString(PWM_KWALLETEMU_DCOPMODULE);
	ok = dcopClient->call("kded", "kded",
			      "loadModule(QCString)",
			      param,
			      replyType,
			      replyData,
			      true);
	if (!ok) {
		printDebug("kded::loadModule(PWM_KWALLETEMU_DCOPMODULE) "
			   "dcop call failed");
		return false;
	}
	if (replyType != "bool") {
		printWarn("kded::loadModule() return "
			  "type != bool");
		return false;
	}
	bool boolRet;
	QDataStream replyStream(replyData, IO_ReadOnly);
	replyStream >> boolRet;
	if (!boolRet) {
		printDebug("loading PWM_KWALLETEMU_DCOPMODULE failed.");
		return false;
	}
	return true;
}

bool KWalletEmu::kdedUnloadKWallet()
{
	bool ok;
	QCString replyType;
	QByteArray replyData;
	ok = dcopClient->call("kded", "kded",
			      "loadedModules()",
			      QByteArray(),
			      replyType,
			      replyData,
			      true);
	if (!ok) {
		printDebug("kded::loadedModules() dcop call failed");
		return false; // kded not running?
	}
	if (replyType != "QCStringList") {
		printWarn("kded::loadedModules() return "
			  "type != QCStringList");
		return false;
	}
	QCStringList modList;
	QDataStream replyStream(replyData, IO_ReadOnly);
	replyStream >> modList;
	if (modList.find("kwalletd") == modList.end()) {
		printDebug("KWalletd module was not loaded.");
		return true;
	}
	QByteArray param;
	QDataStream paramDs(param, IO_WriteOnly);
	paramDs << QCString("kwalletd");
	ok = dcopClient->call("kded", "kded",
			      "unloadModule(QCString)",
			      param,
			      replyType,
			      replyData,
			      true);
	if (!ok) {
		printDebug("kded::unloadModule(QCString) dcop "
			   "call failed");
		return false;
	}
	if (replyType != "bool") {
		printWarn("kded::unloadModule(QCString) return "
			  "type != bool");
		return false;
	}
	bool boolRet;
	replyStream >> boolRet;
	if (!boolRet) {
		printDebug("unloading kwalletd module failed.");
		return false;
	}
	return true;
}

bool KWalletEmu::kdedUnloadEmu()
{
	bool ok;
	QCString replyType;
	QByteArray replyData;
	QByteArray param;
	QDataStream paramDs(param, IO_WriteOnly);
	paramDs << QCString(PWM_KWALLETEMU_DCOPMODULE);
	ok = dcopClient->call("kded", "kded",
			      "unloadModule(QCString)",
			      param,
			      replyType,
			      replyData,
			      true);
	if (!ok) {
		printDebug("kded::unloadModule(PWM_KWALLETEMU_DCOPMODULE) "
			   "dcop call failed");
		return false;
	}
	if (replyType != "bool") {
		printWarn("kded::unloadModule() return "
			  "type != bool");
		return false;
	}
	bool boolRet;
	QDataStream replyStream(replyData, IO_ReadOnly);
	replyStream >> boolRet;
	if (!boolRet) {
		printDebug("unloading PWM_KWALLETEMU_DCOPMODULE failed.");
		return false;
	}
	return true;
}

void KWalletEmu::sig_walletClosed(const QString &wallet)
{
	printDebug("KWalletEmu::sig_walletClosed(const QString&): emitting");
	if (!enabled)
		return;
	QByteArray data;
	QDataStream ds(data, IO_WriteOnly);
	ds << wallet;
	dcopClient->emitDCOPSignal(PWM_KWALLETEMU_DCOPOBJ,
				   "walletClosed(QString)",
				   data);
}

void KWalletEmu::sig_walletClosed(int handle)
{
	printDebug("KWalletEmu::sig_walletClosed(int): emitting");
	if (!enabled)
		return;
	QByteArray data;
	QDataStream ds(data, IO_WriteOnly);
	ds << handle;
	dcopClient->emitDCOPSignal(PWM_KWALLETEMU_DCOPOBJ,
				   "walletClosed(int)",
				   data);
}

void KWalletEmu::sig_allWalletsClosed()
{
	printDebug("KWalletEmu::sig_allWalletsClosed(): emitting");
	if (!enabled)
		return;
	dcopClient->emitDCOPSignal(PWM_KWALLETEMU_DCOPOBJ,
				   "allWalletsClosed()",
				   QByteArray());
}

void KWalletEmu::sig_applicationDisconnected(const QString &wallet,
					     const QCString &application)
{
	printDebug("KWalletEmu::sig_applicationDisconnected(): emitting");
	if (!enabled)
		return;
	QByteArray data;
	QDataStream ds(data, IO_WriteOnly);
	ds << wallet;
	ds << application;
	dcopClient->emitDCOPSignal(PWM_KWALLETEMU_DCOPOBJ,
				   "applicationDisconnected(QString,QCString)",
				   data);
}

void KWalletEmu::sig_folderUpdated(const QString &wallet,
				   const QString &folder)
{
	printDebug("KWalletEmu::sig_folderUpdated(): emitting");
	if (!enabled)
		return;
	QByteArray data;
	QDataStream ds(data, IO_WriteOnly);
	ds << wallet;
	ds << folder;
	dcopClient->emitDCOPSignal(PWM_KWALLETEMU_DCOPOBJ,
				   "folderUpdated(QString,QString)",
				   data);
}

void KWalletEmu::sig_folderListUpdated(const QString &wallet)
{
	printDebug("KWalletEmu::sig_folderListUpdated(): emitting");
	if (!enabled)
		return;
	QByteArray data;
	QDataStream ds(data, IO_WriteOnly);
	ds << wallet;
	dcopClient->emitDCOPSignal(PWM_KWALLETEMU_DCOPOBJ,
				   "folderListUpdated(QString)",
				   data);
}

void KWalletEmu::sig_walletListDirty()
{
	printDebug("KWalletEmu::sig_walletListDirty(): emitting");
	if (!enabled)
		return;
	dcopClient->emitDCOPSignal(PWM_KWALLETEMU_DCOPOBJ,
				   "walletListDirty()",
				   QByteArray());
}

void KWalletEmu::sig_walletCreated(const QString &wallet)
{
	printDebug("KWalletEmu::sig_walletCreated(): emitting");
	if (!enabled)
		return;
	QByteArray data;
	QDataStream ds(data, IO_WriteOnly);
	ds << wallet;
	dcopClient->emitDCOPSignal(PWM_KWALLETEMU_DCOPOBJ,
				   "walletCreated(QString)",
				   data);
}

void KWalletEmu::sig_walletOpened(const QString &wallet)
{
	printDebug("KWalletEmu::sig_walletOpened(): emitting");
	if (!enabled)
		return;
	QByteArray data;
	QDataStream ds(data, IO_WriteOnly);
	ds << wallet;
	dcopClient->emitDCOPSignal(PWM_KWALLETEMU_DCOPOBJ,
				   "walletOpened(QString)",
				   data);
}

void KWalletEmu::sig_walletDeleted(const QString &wallet)
{
	printDebug("KWalletEmu::sig_walletDeleted(): emitting");
	if (!enabled)
		return;
	QByteArray data;
	QDataStream ds(data, IO_WriteOnly);
	ds << wallet;
	dcopClient->emitDCOPSignal(PWM_KWALLETEMU_DCOPOBJ,
				   "walletDeleted(QString)",
				   data);
}

void KWalletEmu::docChanged(PwMDoc *doc)
{
	if (!enabled)
		return;
	sig_folderListUpdated(doc->getTitle());
}

void KWalletEmu::docClosed(PwMDoc *doc)
{
	if (!enabled)
		return;
	sig_walletDeleted(doc->getTitle());
	const std::vector<ConnectionList::connectionElement>
				*cl = connList->getList();
	std::vector<ConnectionList::connectionElement>::const_iterator
				i = cl->begin(),
				end = cl->end();
	while (i != end) {
		if (i->document == doc) {
			doClose(i->handle);
		}
		++i;
	}
}

void KWalletEmu::docOpened(PwMDoc *doc)
{
	if (!enabled)
		return;
	sig_walletCreated(doc->getTitle());
}

void KWalletEmu::docCreated(PwMDoc *doc)
{
	if (!enabled)
		return;
	sig_walletCreated(doc->getTitle());
}

void KWalletEmu::dcopAppDisconn(const QCString &appId)
{
	/* An application disconnected from the DCOP
	 * interface. Make sure there are no dead
	 * connections left.
	 */
	connList->delClient(appId);
}

QStringList KWalletEmu::getOpenDocList() const
{
	QStringList ret;
	const vector<PwMDocList::listItem> *docList = PwMDoc::getOpenDocList()->getList();
	vector<PwMDocList::listItem>::const_iterator dli = docList->begin(),
						     dlend = docList->end();
	while (dli != dlend) {
		ret.append((*dli).docId.c_str());
		++dli;
	}
	return ret;
}

int KWalletEmu::genHandle()
{
	int ret;
	PWM_ASSERT((sizeof(int) == 4) || (sizeof(int) == 8));
	const int intMin = (sizeof(int) == 4) ? INT32_MIN : INT64_MIN;
	Randomizer *rnd = Randomizer::obj();
	do {
		ret = rnd->genRndInt();
		if (ret < 0) {
			if (unlikely(ret == intMin))
				continue;
			ret += 2 * (0 - ret);
		}
	} while (connList->get(ret, 0));
	return ret;
}

int KWalletEmu::connectPeer(const QCString &peer, PwMDoc *doc)
{
	int handle = genHandle();
	connList->add(peer, doc, handle);
	return handle;
}

PwMDoc * KWalletEmu::checkConn(int handle)
{
	if (!enabled)
		return 0;
	ConnectionList::connectionElement e;
	if (!connList->get(handle, &e)) {
		printDebug("KWalletEmu::checkConn(int "
			   "handle):  not found");
		return 0;
	}
	PWM_ASSERT(e.handle == handle);
	if (e.clientName != DCOPPeerName()) {
		printDebug("KWalletEmu::checkConn(int "
			   "handle):  wrong DCOPPeerName()");
		return 0;
	}
	PWM_ASSERT(e.document);
	return e.document;
}

bool KWalletEmu::isEnabled() const
{
	printDebug("KWalletEmu::isEnabled()");
	return enabled;
}

/* open() acts as some kind of "connect to doc" in
 * PwManager. The document has to be opened before this
 * function is called.
 */
int KWalletEmu::open(const QString & wallet, uint /*wId*/)
{
	printDebug("KWalletEmu::open(const QString&, uint)");
	if (!enabled)
		return -1;
	if (wallet.isEmpty())
		return -1;
	QCString appid = DCOPPeerName();
	QStringList docList(getOpenDocList());
	int foundPos = docList.findIndex(wallet);
	if (foundPos == -1)
		return -1; // doc doesn't exist
	const vector <PwMDocList::listItem> *fullDocList = PwMDoc::getOpenDocList()->getList();
	PwMDocList::listItem selectedItem = (*fullDocList)[foundPos];
	if (selectedItem.doc->isDeepLocked()) {
		selectedItem.doc->deepLock(false);
		// a deep-unlock failure is non-fatal, here.
	}
	int handle = connectPeer(appid, selectedItem.doc);
	sig_walletOpened(wallet);
	return handle;
}

int KWalletEmu::openPath(const QString & path, uint /*wId*/)
{
	printDebug("KWalletEmu::openPath(const QString&, uint)");
	if (!enabled)
		return -1;
	if (path.isEmpty())
		return -1;
	QCString appid = DCOPPeerName();
	PwMDoc *doc = mainClass->openDoc(path);
	if (!doc)
		return -1;
	if (doc->isDeepLocked()) {
		doc->deepLock(false);
		// a deep-unlock failure is non-fatal, here.
	}
	int handle = connectPeer(appid, doc);
	sig_walletOpened(doc->getTitle());
	return handle;
}

void KWalletEmu::openAsynchronous(const QString & wallet,
				  const QCString & returnObject,
				  const QCString & returnApp,
				  uint wId)
{
	printDebug("KWalletEmu::openAsynchronous(const QString&, "
		   "const QCString&, const QCString&, uint");
	if (wallet.isEmpty() ||
	    returnObject.isEmpty() ||
	    returnApp.isEmpty()) {
		return;
	}
	int handle = open(wallet, wId);
	QCString appid = DCOPPeerName();
	DCOPRef(appid, "kwalletd").send("walletOpenResult", handle,
					returnObject, returnApp);
}

/* close the connection. _Not_ the document itself */
int KWalletEmu::close(const QString & wallet, bool force)
{
	printDebug("KWalletEmu::close(const QString&, bool)");
	if (!enabled)
		return -1;
	if (wallet.isEmpty())
		return -1;
	QCString appid = DCOPPeerName();
	ConnectionList::connectionElement e;
	if (!connList->find(appid, wallet, &e))
		return -1;
	return close(e.handle, force);
}

/* close the connection. _Not_ the document itself */
int KWalletEmu::close(int handle, bool force)
{
	printDebug("KWalletEmu::close(int, bool)");
	if (!checkConn(handle))
		return -1;
	return doClose(handle, force);
}

int KWalletEmu::doClose(int handle, bool force)
{
	PWM_ASSERT(enabled);
	ConnectionList::connectionElement e;
	if (!connList->get(handle, &e))
		return -1;
	QString docTitle(e.document->getTitle());
	sig_walletClosed(handle);
	sig_walletClosed(docTitle);
	if (force) {
		// close this wallet (document) for all users
		connList->del(docTitle);
	} else {
		// delete only the supplied handle
		connList->del(handle);
	}
	return 0;
}

ASYNC KWalletEmu::sync(int /*handle*/)
{
	printDebug("KWalletEmu::sync(int)");
	/* We do nothing here, because PwManager has
	 * it's own mechanisms to keep the document in sync
	 * with the disk.
	 * This is not true if PwManager crashes, but I
	 * simply assume it doesn't crash. :)
	 */
}

int KWalletEmu::deleteWallet(const QString & wallet)
{
	printDebug("KWalletEmu::deleteWallet(const QString&)");
	if (!enabled)
		return -1;
	if (wallet.isEmpty())
		return -1;
	// we don't delete it here. We just close it.
	QStringList docList(getOpenDocList());
	int foundPos = docList.findIndex(wallet);
	if (foundPos == -1)
		return -1; // doc doesn't exist
	const vector <PwMDocList::listItem> *fullDocList = PwMDoc::getOpenDocList()->getList();
	PwMDocList::listItem selectedItem = (*fullDocList)[foundPos];
	selectedItem.doc->closeDoc();
	return 0;
}

bool KWalletEmu::isOpen(const QString & wallet)
{
	printDebug("KWalletEmu::isOpen(const QString&)");
	if (!enabled)
		return false;
	if (wallet.isEmpty())
		return false;
	return connList->find(wallet, 0);
}

bool KWalletEmu::isOpen(int handle)
{
	printDebug("KWalletEmu::isOpen(int)");
	return checkConn(handle);
}

QStringList KWalletEmu::users(const QString & wallet) const
{
	printDebug("KWalletEmu::users(const QString&)");
	QStringList ret;
	if (!enabled)
		return ret;
	if (wallet.isEmpty())
		return ret;
	const std::vector<ConnectionList::connectionElement> * conn
				= connList->getList();
	std::vector<ConnectionList::connectionElement>::const_iterator
				i = conn->begin(),
				end = conn->end();
	while (i != end) {
		if (wallet == (*i).document->getTitle()) {
			ret.push_back((*i).clientName);
		}
		++i;
	}
	return ret;
}

void KWalletEmu::changePassword(const QString & wallet, uint /*wId*/)
{
	printDebug("KWalletEmu::changePassword(const QString&, uint)");
	if (!enabled)
		return;
	if (wallet.isEmpty())
		return;
	QStringList docList(getOpenDocList());
	int foundPos = docList.findIndex(wallet);
	if (foundPos == -1)
		return; // doc doesn't exist
	const vector <PwMDocList::listItem> *fullDocList = PwMDoc::getOpenDocList()->getList();
	PwMDocList::listItem selectedItem = (*fullDocList)[foundPos];
	selectedItem.doc->changeCurrentPw();
}

QStringList KWalletEmu::wallets() const
{
	printDebug("KWalletEmu::wallets()");
	if (!enabled)
		return QStringList();
	return getOpenDocList();
}

QStringList KWalletEmu::folderList(int handle)
{
	printDebug("KWalletEmu::folderList(int)");
	QStringList ret;
	PwMDoc *doc = checkConn(handle);
	if (!doc)
		return ret;
	doc->getCategoryList(&ret);
	return ret;
}

bool KWalletEmu::hasFolder(int handle, const QString &folder)
{
	printDebug("KWalletEmu::hasFolder(int, const QString&)");
	if (folder.isEmpty()) {
		// we never have an empty category name.
		return false;
	}
	PwMDoc *doc = checkConn(handle);
	if (!doc)
		return false;
	const char *_folder = folder.latin1();
	vector<string> catList;
	doc->getCategoryList(&catList);
	vector<string>::iterator i = catList.begin(),
				 end = catList.end();
	while (i != end) {
		if (*i == _folder)
			return true;
		++i;
	}
	return false;
}

bool KWalletEmu::createFolder(int handle, const QString &folder)
{
	printDebug("KWalletEmu::createFolder(int, const QString&)");
	if (folder.isEmpty())
		return false;
	PwMDoc *doc = checkConn(handle);
	if (!doc)
		return false;
	PwMerror err = doc->addCategory(folder, 0, true);
	if (err != e_success)
		return false;
	doc->flagDirty(); // we have to do this manually
	return true;
}

bool KWalletEmu::removeFolder(int handle, const QString & folder)
{
	printDebug("KWalletEmu::removeFolder(int, const QString&)");
	if (folder.isEmpty())
		return false;
	PwMDoc *doc = checkConn(handle);
	if (!doc)
		return false;
	return doc->delCategory(folder);
}

QStringList KWalletEmu::entryList(int handle, const QString & folder)
{
	printDebug("KWalletEmu::entryList(int, const QString&)");
	QStringList ret;
	if (folder.isEmpty())
		return ret;
	PwMDoc *doc = checkConn(handle);
	if (!doc)
		return ret;
	doc->getEntryList(folder, &ret);
	return ret;
}

QByteArray KWalletEmu::readEntry(int /*handle*/, const QString & /*folder*/,
				 const QString & /*key*/)
{
	printDebug("KWalletEmu::readEntry(int, const QString&, const QString&)");
	TODO(); // TODO
	printInfo("KWalletEmu::readEntry(int, const QString&, const QString&): "
		  "This feature is currently not supported.");
	return QByteArray();
}

QByteArray KWalletEmu::readMap(int /*handle*/, const QString & /*folder*/,
			       const QString & /*key*/)
{
	printDebug("KWalletEmu::readMap(int, const QString&, const QString&)");
	TODO(); // TODO
	printInfo("KWalletEmu::readMap(int, const QString&, const QString&): "
		  "This feature is currently not supported.");
	return QByteArray();
}

QString KWalletEmu::readPassword(int handle, const QString & folder,
				 const QString & key)
{
	printDebug("KWalletEmu::readPassword(int, const QString&, const QString&)");
	PwMDoc *doc = checkConn(handle);
	if (!doc)
		return "";
	if (doc->isDeepLocked()) {
		PwMerror err = doc->deepLock(false);
		if (err != e_success)
			return "";
	}
	if (folder.isEmpty() || key.isEmpty())
		return "";
	vector<unsigned int> foundPos;
	PwMDataItem d;
	d.desc = key.latin1();
	doc->findEntry(folder, d, SEARCH_IN_DESC,
		       &foundPos, true);
	if (!foundPos.size())
		return "";
	if (!doc->getEntry(folder, foundPos[0], &d, true))
		return "";
	return d.pw.c_str();
}

int KWalletEmu::renameEntry(int handle, const QString & folder,
			    const QString & oldName, const QString & newName)
{
	printDebug("KWalletEmu::renameEntry(int, const QString&, const QString&, const QString&)");
	PwMDoc *doc = checkConn(handle);
	if (!doc)
		return -1;
	if (folder.isEmpty() ||
	    oldName.isEmpty() ||
	    newName.isEmpty()) {
		return -1;
	}
	vector<unsigned int> foundPos;
	PwMDataItem d;
	d.desc = oldName.latin1();
	doc->findEntry(folder, d, SEARCH_IN_DESC,
		       &foundPos, true);
	if (!foundPos.size())
		return -1;
	if (!doc->getEntry(folder, foundPos[0], &d, true))
		return -1;
	d.desc = newName.latin1();
	if (!doc->editEntry(folder, folder, foundPos[0], &d))
		return -1;
	return 0;
}

int KWalletEmu::writeEntry(int handle, const QString & folder,
			   const QString & key, const QByteArray & value,
			   int entryType)
{
	printDebug("KWalletEmu::writeEntry(int, const QString&, const QString&, const QByteArray&, int)");
	if (entryType != Password) {
		TODO(); // TODO
		printInfo("KWalletEmu::writeEntry(int, const QString&, "
			  "const QString&, const QByteArray&, int): "
			  "This entryType is currently not supported.");
		// other types are currently not supported.
		return -1;
	}
	if (folder.isEmpty() ||
	    key.isEmpty() ||
	    value.isEmpty()) {
		return -1;
	}
	QString strval;
	QDataStream ds(value, IO_ReadOnly);
	ds >> strval;
	return writePassword(handle, folder, key, strval);
}

int KWalletEmu::writeEntry(int /*handle*/, const QString & /*folder*/,
			   const QString & /*key*/, const QByteArray & /*value*/)
{
	printDebug("KWalletEmu::writeEntry(int, const QString&, const QString&, const QByteArray&)");
	TODO(); // TODO
	printInfo("KWalletEmu::writeEntry(int, const QString&, const QString&, const QByteArray&): "
		  "This feature is currently not supported.");
	// not supported (yet?)
	return -1;
}

int KWalletEmu::writeMap(int /*handle*/, const QString & /*folder*/,
			 const QString & /*key*/, const QByteArray & /*value*/)
{
	printDebug("KWalletEmu::writeMap(int, const QString&, const QString&, const QByteArray&)");
	TODO(); // TODO
	printInfo("KWalletEmu::writeMap(int, const QString&, const QString&, const QByteArray&): "
		  "This feature is currently not supported.");
	// not supported (yet?)
	return -1;
}

int KWalletEmu::writePassword(int handle, const QString & folder,
			      const QString & key, QString value)
{
	printDebug("KWalletEmu::writePassword(int, const QString&, const QString&, const QString&)");
	PwMDoc *doc = checkConn(handle);
	if (!doc)
		return -1;
	if (folder.isEmpty() ||
	    key.isEmpty()) {
		return -1;
	}
	if (value.isEmpty())
		value = "";
	bool ok;
	PwMerror err;
	vector<unsigned int> foundPos;
	PwMDataItem d;
	d.clear();
	d.desc = key.latin1();
	doc->findEntry(folder, d, SEARCH_IN_DESC,
		       &foundPos, true);
	if (foundPos.size()) {
		/* the entry does already exist.
		 * edit it.
		 */
		printDebug("KWalletEmu::writePassword(): edit existing");
		ok = doc->getEntry(folder, foundPos[0], &d, true);
		if (!ok)
			return -1;
		d.pw = value.latin1();
		ok = doc->editEntry(folder, folder,
				    foundPos[0], &d);
		if (!ok)
			return -1;
		return 0;
	}
	/* the entry does not exist.
	 * create a new one.
	 */
	printDebug("KWalletEmu::writePassword(): create new");
	d.pw = value.latin1();
	err = doc->addEntry(folder, &d);
	if (err != e_success)
		return -1;
	return 0;
}

bool KWalletEmu::hasEntry(int handle, const QString & folder,
			  const QString & key)
{
	printDebug("KWalletEmu::hasEntry(int, const QString&, const QString&)");
	PwMDoc *doc = checkConn(handle);
	if (!doc)
		return false;
	if (folder.isEmpty() ||
	    key.isEmpty()) {
		return false;
	}
	vector<unsigned int> foundPos;
	PwMDataItem d;
	d.desc = key.latin1();
	doc->findEntry(folder, d, SEARCH_IN_DESC,
		       &foundPos, true);
	if (!foundPos.size())
		return false;
	return true;
}

int KWalletEmu::entryType(int handle, const QString & /*folder*/,
			  const QString & /*key*/)
{
	printDebug("KWalletEmu::entryType(int, const QString&, const QString&)");
	PwMDoc *doc = checkConn(handle);
	if (!doc)
		return Unknown;
	// TODO: If we support maps and streams someday we have to fix this up.
	return Password;
}

int KWalletEmu::removeEntry(int handle, const QString & folder,
			    const QString & key)
{
	printDebug("KWalletEmu::removeEntry(int, const QString&, const QString&)");
	PwMDoc *doc = checkConn(handle);
	if (!doc)
		return false;
	if (folder.isEmpty() ||
	    key.isEmpty()) {
		return false;
	}
	vector<unsigned int> foundPos;
	PwMDataItem d;
	d.desc = key.latin1();
	doc->findEntry(folder, d, SEARCH_IN_DESC,
		       &foundPos, true);
	if (!foundPos.size())
		return -1;
	if (!doc->delEntry(folder, foundPos[0]))
		return -1;
	return 0;
}

bool KWalletEmu::disconnectApplication(const QString & wallet,
				       const QCString & application)
{
	printDebug("KWalletEmu::disconnectApplication(const QString&, const QCString&)");
	if (!enabled)
		return false;
	if (wallet.isEmpty() ||
	    application.isEmpty()) {
		return false;
	}
	ConnectionList::connectionElement element;
	if (!connList->find(application, wallet, &element))
		return false;
	connList->del(element.handle);
	sig_applicationDisconnected(wallet, application);
	return true;
}

void KWalletEmu::reconfigure()
{
	printDebug("KWalletEmu::reconfigure()");
	// we don't do anything here, yet.
}

bool KWalletEmu::folderDoesNotExist(const QString & wallet,
				    const QString & folder)
{
	printDebug("KWalletEmu::folderDoesNotExist(const QString&, const QString&)");
	if (!enabled)
		return true;
	if (wallet.isEmpty() ||
	    folder.isEmpty()) {
		return true;
	}
	QStringList docList(getOpenDocList());
	int foundPos = docList.findIndex(wallet);
	if (foundPos == -1)
		return true; // doc doesn't exist
	const vector <PwMDocList::listItem> *fullDocList = PwMDoc::getOpenDocList()->getList();
	PwMDocList::listItem selectedItem = (*fullDocList)[foundPos];
	if (selectedItem.doc->findCategory(folder, 0))
		return false;
	return true;
}

bool KWalletEmu::keyDoesNotExist(const QString & wallet, const QString & folder,
				 const QString & key)
{
	printDebug("KWalletEmu::keyDoesNotExist(const QString&, const QString&, const QString&)");
	if (!enabled)
		return true;
	if (wallet.isEmpty() ||
	    folder.isEmpty() ||
	    key.isEmpty()) {
		return true;
	}
	QStringList docList(getOpenDocList());
	int foundPos = docList.findIndex(wallet);
	if (foundPos == -1)
		return true; // doc doesn't exist
	const vector <PwMDocList::listItem> *fullDocList = PwMDoc::getOpenDocList()->getList();
	PwMDocList::listItem selectedItem = (*fullDocList)[foundPos];
	vector<unsigned int> found;
	PwMDataItem d;
	d.desc = key.latin1();
	selectedItem.doc->findEntry(folder, d, SEARCH_IN_DESC,
				    &found, true);
	if (found.size())
		return false;
	return true;
}

void KWalletEmu::closeAllWallets()
{
	printDebug("KWalletEmu::closeAllWallets()");
	if (!enabled)
		return;
	const vector<ConnectionList::connectionElement> *cl = connList->getList();
	vector<ConnectionList::connectionElement>::const_iterator i = cl->begin(),
								  end = cl->end();
	bool cleanClose = true;
	while (i != end) {
		if (doClose(i->handle, true))
			cleanClose = false;
		++i;
	}
	if (cleanClose)
		sig_allWalletsClosed();
}

#include "kwalletemu.moc"
