/***************************************************************************
 *                                                                         *
 *   copyright (C) 2004 by Michael Buesch                                  *
 *   email: mbuesch@freenet.de                                             *
 *                                                                         *
 *   KWallet DCOP Interface is                                             *
 *    Copyright (c) 2002-2003 George Staikos <staikos@kde.org>             *
 *   KWallet is part of the KDE libraries                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License version 2        *
 *   as published by the Free Software Foundation.                         *
 *                                                                         *
 ***************************************************************************/

#ifndef KWALLETEMU_H
#define KWALLETEMU_H

#include <qstring.h>
#include <qcstring.h>
#include <qstringlist.h>

#include <dcopobject.h>
#include <dcopclient.h>

#include <vector>

class PwMDoc;
class PwM;

/** This is a class that emulates the KWallet
  * DCOP interface for PwManager, so you can use
  * PwManager for KWallet-tasks.
  * Yea, I know I'm a stupid, ignorant programmer and
  * I should go to hell for this piece of software.
  * But it's "just for fun". :)
  * Please forgive me.
  *
  * Important note: This emulation assumes, that each entry
  *                 can be identified by description. This is
  *                 _not_ PwManager behaviour, so it may
  *                 lead to problems if you have many entries
  *                 with same desc.
  *
  * List of all DCOP-signals this class emits:
  *   walletClosed(QString wallet)
  *   walletClosed(int handle)
  *   allWalletsClosed()
  *   applicationDisconnected(QString wallet, QCString application)
  *   folderUpdated(QString wallet, QString folder)
  *   folderListUpdated(QString wallet)
  *   walletListDirty()
  *   walletCreated(QString wallet)
  *   walletOpened(QString wallet)
  *   walletDeleted(QString wallet)
  */
class KWalletEmu : public QObject
		 , public DCOPObject
{
	Q_OBJECT
	K_DCOP

protected:
	class ConnectionList
	{
	public:
		struct connectionElement
		{
			QCString	clientName;
			PwMDoc		*document;
			int		handle;
		};

	public:
		ConnectionList(KWalletEmu *_kwalletEmu)
				{ kwalletEmu = _kwalletEmu; }

		void add(const QCString &client, PwMDoc *doc, int handle);
		/** delete the connection with this handle */
		void del(int handle);
		/** delete all connections to this wallet */
		void del(const QString &wallet);
		/** delete all connections from this client */
		void delClient(const QCString &client);
		bool get(int handle, connectionElement *ret);
		bool find(const QCString &client, const QString &wallet,
			  connectionElement *ret);
		/** find a connection to "wallet". Returns the first
		  * element that has the name "wallet"
		  */
		bool find(const QString wallet, connectionElement *ret);

		const std::vector<connectionElement> * getList()
				{ return &connList; }
	
	protected:
		std::vector<connectionElement> connList;
		KWalletEmu *kwalletEmu;
	};

	enum EntryType
	{
		Unknown = 0,
		Password,
		Stream,
		Map,
		Unused=0xffff
	};

public:
	friend class KWalletEmu::ConnectionList;

	KWalletEmu(const QCString &objId, QObject *parent,
		   PwM *_mainClass, const char *name = 0);
	virtual ~KWalletEmu();

	/** enable/disable the KWalletEmu interface */
	void setEnabled(bool _enabled)
			{ enabled = _enabled; }
	/** is the KWalletEmu interface enabled? */
	bool isEnabled()
			{ return enabled; }
	/** connects all neccessary signals from "doc" to the
	  * emulation layer. Call this on every new document creation
	  */
	void connectDocSignals(PwMDoc *doc);

protected:
	/** get the name of the current peer */
	QCString DCOPPeerName();
	/** get a list of all open document-names (titles) */
	QStringList getOpenDocList() const;
	/** generate a new random and unique handle */
	int genHandle();
	/** check if this peer has a connection and
	  * return its document pointer
	  */
	PwMDoc * checkConn(int handle);
	/** make a new connection entry for this peer */
	int connectPeer(const QCString &peer, PwMDoc *doc);
	/** load the real KWallet KDED module.
	  * It only loads it if KWallet is enabled in its config
	  */
	int kdedLoadKWallet();
	/** load the KWallet KDED emulation module */
	bool kdedLoadEmu();
	/** unload the real KWallet KDED module */
	bool kdedUnloadKWallet();
	/** unload the KWallet KDED emulation module */
	bool kdedUnloadEmu();
	/** close the document connection with the supplied handle.
	  * This function does _not_ call checkConn()
	  */
	int doClose(int handle, bool force = false);

protected slots:
	/** the document "doc" changed its data */
	void docChanged(PwMDoc *doc);
	/** the document "doc" got closed */
	void docClosed(PwMDoc *doc);
	/** the document "doc" got just opened */
	void docOpened(PwMDoc *doc);
	/** the document "doc" got created */
	void docCreated(PwMDoc *doc);
	/** an application disconnected from the dcop interface */
	void dcopAppDisconn(const QCString &appId);

protected:
	/* signal sending methods */
	void sig_walletClosed(const QString &wallet);
	void sig_walletClosed(int handle);
	void sig_allWalletsClosed();
	void sig_applicationDisconnected(const QString &wallet,
					 const QCString &application);
	void sig_folderUpdated(const QString &wallet,
			       const QString &folder);
	void sig_folderListUpdated(const QString &wallet);
	void sig_walletListDirty();
	void sig_walletCreated(const QString &wallet);
	void sig_walletOpened(const QString &wallet);
	void sig_walletDeleted(const QString &wallet);

protected:
	/** Interface enabled? */
	bool enabled;
	/** list of all open connections */
	ConnectionList *connList;
	/** pointer to the main application class */
	PwM *mainClass;

k_dcop:
	virtual bool isEnabled() const;
	virtual int open(const QString& wallet, uint wId);
	virtual int openPath(const QString& path, uint wId);
	virtual void openAsynchronous(const QString& wallet,
				      const QCString& returnObject,
				      const QCString& returnApp,
				      uint wId);
	virtual int close(const QString& wallet, bool force);
	virtual int close(int handle, bool force);
	virtual ASYNC sync(int handle);
	virtual int deleteWallet(const QString& wallet);
	virtual bool isOpen(const QString& wallet);
	virtual bool isOpen(int handle);
	virtual QStringList users(const QString& wallet) const;
	virtual void changePassword(const QString& wallet,
				    uint wId);
	virtual QStringList wallets() const;
	virtual QStringList folderList(int handle);
	virtual bool hasFolder(int handle,
			       const QString &folder);
	virtual bool createFolder(int handle,
				  const QString& folder);
	virtual bool removeFolder(int handle,
				  const QString& folder);
	virtual QStringList entryList(int handle,
				      const QString& folder);
	virtual QByteArray readEntry(int handle,
				     const QString& folder,
				     const QString& key);
	virtual QByteArray readMap(int handle,
				   const QString& folder,
				   const QString& key);
	virtual QString readPassword(int handle,
				     const QString& folder,
				     const QString& key);
	virtual int renameEntry(int handle,
				const QString& folder,
				const QString& oldName,
				const QString& newName);
	virtual int writeEntry(int handle,
			       const QString& folder,
			       const QString& key,
			       const QByteArray& value,
			       int entryType);
	virtual int writeEntry(int handle,
			       const QString& folder,
			       const QString& key,
			       const QByteArray& value);
	virtual int writeMap(int handle,
			     const QString& folder,
			     const QString& key,
			     const QByteArray& value);
	virtual int writePassword(int handle,
				  const QString& folder,
				  const QString& key,
				  QString value);
	virtual bool hasEntry(int handle,
			      const QString& folder,
			      const QString& key);
	virtual int entryType(int handle,
			      const QString& folder,
			      const QString& key);
	virtual int removeEntry(int handle,
				const QString& folder,
				const QString& key);
	virtual bool disconnectApplication(const QString& wallet,
					   const QCString& application);
	virtual void reconfigure();
	virtual bool folderDoesNotExist(const QString& wallet,
					const QString& folder);
	virtual bool keyDoesNotExist(const QString& wallet,
				     const QString& folder,
				     const QString& key);
	virtual void closeAllWallets();
};

#endif
