/***************************************************************************
 *                                                                         *
 *   copyright (C) 2003, 2004 by Michael Buesch                            *
 *   email: mbuesch@freenet.de                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License version 2        *
 *   as published by the Free Software Foundation.                         *
 *                                                                         *
 ***************************************************************************/

#ifndef PWM_H
#define PWM_H

#include <kapp.h>
#include <klistview.h>
#include <kmainwindow.h>
#include <kpopupmenu.h>
#include <kprocess.h>
#include <kaction.h>
#include <kdeversion.h>
#include <kwin.h>

#include <qworkspace.h>
#include <qglobal.h>

#include <vector>

#include "pwmview.h"
#include "pwmtray.h"
#include "pwmexception.h"

#define PWM_DEFAULT_DCOPOBJ		"pwmanager"
#define PWM_KWALLETEMU_DCOPMODULE	"pwmanager_kwalletemu"
#define PWM_KWALLETEMU_DCOPOBJ		"kwalletemu"
/** timeout for displaying a message on the status-bar (in seconds) */
#define STATUSBAR_MSG_TIMEOUT	5


ConfFile* getConfFile();

class PwMKeyCard;
class PwManagerDCOPif;
class KWalletEmu;

using std::vector;

/** this is a list of all open view widgets */
class ViewWidgetList
{
public:
	ViewWidgetList() {}

	void add(PwMView *view)
		{
# ifdef PWM_DEBUG
			vector<PwMView *>::iterator i = viewList.begin(),
						    end = viewList.end();
			while (i != end) {
				if (*i == view) {
					BUG();
					return;
				}
				++i;
			}
# endif // PWM_DEBUG
			viewList.push_back(view);
		}
	void del(PwMView *view)
		{
			vector<PwMView *>::iterator i = viewList.begin(),
						    end = viewList.end();
			while (i != end) {
				if (*i == view) {
					viewList.erase(i);
					return;
				}
				++i;
			}
		}
	const vector<PwMView *> * getList()
			{ return &viewList; }

protected:
	vector<PwMView *> viewList;
};

/** PwM is the base class of the project */
class PwM : public KMainWindow
{
	Q_OBJECT
public:
	/** construtor */
	PwM(QWidget* parent = 0, const char *name = 0);
	/** destructor */
	~PwM();

	/** copy some text to the global clipboard */
	static void copyToClipboard(const QString &s);

	/** returns pointer to current MDI-Window */
	PwMView* curView()
			{ return currentView; }
	/** returns pointer to the currently using document. */
	PwMDoc* curDoc()
			{ PWM_ASSERT(curView()); return curView()->document(); }
	/** show/create the tray-symbol */
	void showTray();
	/** hide/delete the tray-symbol */
	void hideTray();
	/** make all connections for the tray icon */
	void connectTray(PwMDoc *doc);
	/** is the tray icon present/visible? */
	bool isTrayVisible()
			{ return (tray ? true : false); }
	/** open a new doc with the given filename */
	PwMDoc * openDoc(QString filename, bool openDeepLocked = false);
	/** show a message on the global status bar.
	  * The message times out after some seconds.
	  */
	void showStatMsg(const QString &msg);

public slots:
	/** handle user changed MDI window */
	void wndActivated_slot(QWidget *w);
	/** file/new triggered */
	void new_slot();
	/** file/open triggered */
	void open_slot(QString fn = "");
	/** file/close triggered */
	void close_slot();
	/** file/save triggered */
	void save_slot();
	/** file/saveAs triggered */
	void saveAs_slot();
	/** file/export/text triggered */
	void exportToText();
	/** file/export/gpasman triggered */
	void exportToGpasman();
	/** file/export/kwallet triggered */
	void exportToKWallet();
	/** file/import/text triggered */
	void importFromText();
	/** file/import/gpasman triggered */
	void importFromGpasman();
	/** file/import/kwallet triggered */
	void importKWallet();
	/** file/print triggered */
	void print_slot();
	/** manage/add triggered */
	void addPwd_slot(QString *pw = 0, PwMDoc *_doc = 0);
	/** manage/edit triggered */
	void editPwd_slot(const QString *category = 0, const int *index = 0,
			  PwMDoc *_doc = 0);
	/** manage/delete triggered */
	void deletePwd_slot();
	/** execute the "Launcher" entry */
	void execLauncher_slot();
	/** open browser with URL entry */
	void goToURL_slot();
	/** manage/changeMasterPwd triggered */
	void changeMasterPwd_slot();
	/** lock current document */
	void lockWnd_slot();
	/** lock all open documents */
	void lockAllDoc_slot();
	/** deeplock current document */
	void deepLockWnd_slot();
	/** deeplock all open documents */
	void deepLockAllDoc_slot();
	/** window/unlock triggered */
	void unlockWnd_slot();
	/** find item */
	void find_slot();
	/** enable/disable mdi-manage-user-interface */
	void enableMdiUi(bool enable = true);
	/** configure clicked */
	void config_slot();
	/** unlock triggered from tray */
	void trayUndockButton_slot();
	/** (de)activate the "change master pw" button in the menu-bar */
	void activateMpButton(bool activate = true);
	/** start PwG is installed */
	void startPwG_slot();
	/** generate a new chipcard */
	void genNewCard_slot();
	/** completely erase the current card */
	void eraseCard_slot();
	/** returns the ID number of the current card */
	void readCardId_slot();
	/** displays the "keycard-support not available" msg-box */
	void noCard_slot();
	/** make backup image of the current card */
	void makeCardBackup_slot();
	/** write backup image to current card */
	void replayCardBackup_slot();
	/** update the "window" popup menu */
	void updateWindowPopup();
	/** the user clicked a button in the "window" popup menu */
	void selectWindowPressed(int id);

protected slots:
	/** an MDI PwMView widget got closed */
	void viewClosed(PwMView *view);

protected:
	/** initialize the menubar */
	void initMenubar();
	/** initialize the toolbar */
	void initToolbar();
	/** initialize the window-metrics */
	void initMetrics();
	/** read the background-pixmap from config file and display it */
	void initBackgroundPix();
	/** initialize the dcop interface */
	void initDCOP();
	/** initialize the KWallet emulation */
#ifdef CONFIG_KWALLETIF
	void initKWalletEmu();
#else
	void initKWalletEmu() {}
#endif // CONFIG_KWALLETIF
	/** close-event */
	void closeEvent(QCloseEvent *e);
	/** creates a new PwM-ListView and returns it */
	PwMView* makeNewListView();
	/** creates a new document */
	PwMDoc* makeNewDocument();
	/** Window hide-event */
	void hideEvent(QHideEvent *);
	/** is this window minimized? */
	bool isMinimized()
		{
		  #if KDE_VERSION >= KDE_MAKE_VERSION(3, 2, 0)
			return KWin::windowInfo(winId()).isMinimized();
		  #else // KDE_VERSION
			return KWin::info(winId()).isIconified();
		  #endif // KDE_VERSION
		}
	QWidgetList getMdiWidgetList()
		{
			PWM_ASSERT(workspace);
		  #if QT_VERSION >= QT_MAKE_VERSION(3, 2, 0)
			return workspace->windowList(QWorkspace::CreationOrder);
		  #else // QT_VERSION
			return workspace->windowList();
		  #endif // QT_VERSION
		}

protected:
	/** MDI workspace */
	QWorkspace *workspace;
	/** active MDI window.
	  * It's updated by wndActivated_slot() , so
	  * it should give always the current state.
	  * Note the "should". :)
	  */
	PwMView *currentView;
	/** list of all open MDI views */
	ViewWidgetList vl;
	/** "file" popup-menu */
	KPopupMenu *filePopup;
	/** "manage" popup-menu */
	KPopupMenu *managePopup;
	/** "chipcard" popup-menu */
	KPopupMenu *chipcardPopup;
	/** "window" popup-menu */
	KPopupMenu *windowPopup;
	/** "view" popup-menu */
	KPopupMenu *viewPopup;
	/** "options" popup-menu */
	KPopupMenu *optionsPopup;
	/** "help" popup-menu */
	KPopupMenu *helpPopup;
	/** "export" popup-menu */
	KPopupMenu *exportPopup;
	/** "import" popup-menu */
	KPopupMenu *importPopup;
	/** recent files action */
	KRecentFilesAction *actRecentFiles;
	/** tray-icon */
	PwMTray *tray;
	/** PwG process object */
	KProcess pwgproc;
	/** current background pixmap */
	QPixmap *currBg;
	/** keycard access class */
#ifdef CONFIG_KEYCARD
	PwMKeyCard *keyCard;
#endif
#ifdef CONFIG_KWALLETIF
	KWalletEmu *kwalletEmu;
#endif
	/** DCOP-interface object */
	PwManagerDCOPif *dcopIf;
};

#endif
