/***************************************************************************
 *                                                                         *
 *   copyright (C) 2003, 2004 by Michael Buesch                            *
 *   email: mbuesch@freenet.de                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License version 2        *
 *   as published by the Free Software Foundation.                         *
 *                                                                         *
 ***************************************************************************/

#include "pwmdoc.h"
#include "pwmview.h"
#include "blowfish.h"
#include "sha1.h"
#include "globalstuff.h"
#include "gpasmanfile.h"
#include "serializer.h"
#include "compressgzip.h"
#include "compressbzip2.h"
#include "pwmanagerdcopif.h"

#include <qdatetime.h>
#include <qsize.h>
#include <qfileinfo.h>
#include <qfile.h>

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <iostream>
#include <algorithm>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdint.h>

using namespace std;


void PwMDocList::add(PwMDoc *doc, const string &id)
{
#ifdef PWM_DEBUG
	// check for existance of object in debug mode only.
	vector<listItem>::iterator begin = docList.begin(),
				   end = docList.end(),
				   i = begin;
	while (i != end) {
		if (i->doc == doc) {
			BUG();
			return;
		}
		++i;
	}
#endif
	listItem newItem;
	newItem.doc = doc;
	newItem.docId = id;
	docList.push_back(newItem);
}

void PwMDocList::edit(PwMDoc *doc, const string &newId)
{
	vector<listItem>::iterator begin = docList.begin(),
				   end = docList.end(),
				   i = begin;
	while (i != end) {
		if (i->doc == doc) {
			i->docId = newId;
			return;
		}
		++i;
	}
}

void PwMDocList::del(PwMDoc *doc)
{
	vector<listItem>::iterator begin = docList.begin(),
				   end = docList.end(),
				   i = begin;
	while (i != end) {
		if (i->doc == doc) {
			docList.erase(i);
			return;
		}
		++i;
	}
}

bool PwMDocList::find(const string &id, listItem *ret)
{
	vector<listItem>::iterator begin = docList.begin(),
				   end = docList.end(),
				   i = begin;
	while (i != end) {
		if (i->docId == id) {
			if (ret)
				*ret = *i;
			return true;
		}
		++i;
	}
	return false;
}

PwMDocList PwMDoc::openDocList;
unsigned int PwMDocList::unnamedDocCnt = 1;

PwMDoc::PwMDoc(QObject *parent, const char *name)
 : PwMDocUi(parent, name)
{
	unnamedNum = 0;
	getOpenDocList()->add(this, getTitle().latin1());
	curDocStat = 0;
	setMaxNumEntries();
	mpwTimer = new QTimer(this);
	autoLockTimer = new QTimer(this);
	setMpwTimeout();
	setAutoLockTimeout(conf()->confGlobLockTimeout());
	unsigned int cat;
	addCategory(DEFAULT_CATEGORY, &cat, false);
	listView = 0;
	connect(mpwTimer, SIGNAL(timeout()), this, SLOT(resetUnlockWoPw()));
	connect(autoLockTimer, SIGNAL(timeout()), this, SLOT(autoLockDocument()));
	emit docCreated(this);
}

PwMDoc::~PwMDoc()
{
	emit docClosed(this);
	if (getDcopConnFlag() && pwmDCOPinterface) {
		pwmDCOPinterface->closeConn(this);
	}
	getOpenDocList()->del(this);
}

void PwMDoc::closeDoc()
{
	delete this;
}

PwMerror PwMDoc::saveDoc(char compress, const QString *file)
{
	if (!file) {
		if (filename == "")
			return e_filename;
	} else {
		if (*file == "" && filename == "")
			return e_filename;
		if (*file != "")
			filename = *file;
	}

	bool wasDeepLocked = isDeepLocked();
	if (wasDeepLocked) {
		if (deepLock(false) != e_success)
			return e_noPw;
	}

	if (!isPwAvailable()) {
		/* password is not available. This means, the
		 * document wasn't saved, yet.
		 */
		bool useChipcard = getDocStatFlag(DOC_STAT_USE_CHIPCARD);
		QString pw(requestNewMpw(&useChipcard));
		if (pw != "") {
			currentPw = pw;
		} else {
			return e_noPw;
		}
		if (useChipcard) {
			setDocStatFlag(DOC_STAT_USE_CHIPCARD);
		} else {
			unsetDocStatFlag(DOC_STAT_USE_CHIPCARD);
		}
	}

	if (conf()->confGlobMakeFileBackup()) {
		if (!backupFile(filename))
			return e_fileBackup;
	}
	if (QFile::exists(filename))
		QFile::remove(filename);
	QFile f(filename);
	if (!f.open(IO_ReadWrite))
		return e_openFile;

	if (writeFileHeader(PWM_HASH_SHA1, PWM_HASH_SHA1,
			    PWM_CRYPT_BLOWFISH, compress,
			    &currentPw, &f) != e_success) {
		printDebug("PwMDoc::saveDoc(): writeFileHeader() failed");
		f.close();
		return e_writeHeader;
	}
	string serialized;
	if (!serializeDta(&serialized)) {
		printDebug("PwMDoc::saveDoc(): serializeDta() failed");
		f.close();
		return e_serializeDta;
	}
	if (writeDataHash(PWM_HASH_SHA1, &serialized, &f) != e_success) {
		printDebug("PwMDoc::saveDoc(): writeDataHash() failed");
		f.close();
		return e_writeHeader;
	}
	if (!compressDta(&serialized, compress)) {
		printDebug("PwMDoc::saveDoc(): compressDta() failed");
		f.close();
		return e_enc;
	}
	PwMerror ret = encrypt(&serialized, &currentPw, &f, PWM_CRYPT_BLOWFISH);
	if (ret == e_weakPw) {
		printDebug("PwMDoc::saveDoc(): encrypt() failed");
		f.close();
		return e_weakPw;
	} else if (ret != e_success) {
		printDebug("PwMDoc::saveDoc(): encrypt() failed");
		f.close();
		return e_enc;
	}
	unsetDocStatFlag(DOC_STAT_DISK_DIRTY);
	f.close();
	if (chmod(filename.latin1(),
		  conf()->confGlobFilePermissions())) {
		printWarn(string("chmod failed: ") + strerror(errno));
	}
	openDocList.edit(this, getTitle().latin1());
	if (wasDeepLocked)
		deepLock(true);

	return e_success;
}

PwMerror PwMDoc::openDoc(const QString *file, int openLocked)
{
	PWM_ASSERT(file);
	PWM_ASSERT(openLocked == 0 || openLocked == 1 || openLocked == 2);
	string decrypted, dataHash;
	PwMerror ret;
	char cryptAlgo, dataHashType, compress;
	unsigned int headerLen;

	if (*file == "")
		return e_readFile;
	filename = *file;
	/* check if this file is already open.
	 * This does not catch symlinks!
	 */
	if (!isDeepLocked()) {
		if (getOpenDocList()->find(filename.latin1()))
			return e_alreadyOpen;
	}
	QFile f(filename);

	if (openLocked == 2) {
		// open deep-locked
		if (!QFile::exists(filename))
			return e_openFile;
		if (deepLock(true, false) != e_success)
			return e_openFile;
		goto out_success;
	}

	if (!f.open(IO_ReadOnly))
		return e_openFile;

	ret = checkHeader(&cryptAlgo, &currentPw, &compress, &headerLen,
			  &dataHashType, &dataHash, &f);
	if (ret != e_success) {
		printDebug("PwMDoc::openDoc(): checkHeader() failed");
		f.close();
		if (ret == e_wrongPw) {
			wrongMpwMsgBox(getDocStatFlag(DOC_STAT_USE_CHIPCARD));
			return ret;
		} else if (ret == e_noPw ||
			   ret == e_fileVer ||
			   ret == e_fileFormat) {
			return ret;
		} else
			return e_readFile;
	}
	if (decrypt(&decrypted, headerLen, &currentPw, cryptAlgo, &f) != e_success) {
		printDebug("PwMDoc::openDoc(): decrypt() failed");
		f.close();
		return e_readFile;
	}
	if (!decompressDta(&decrypted, compress)) {
		printDebug("PwMDoc::openDoc(): decompressDta() failed");
		f.close();
		return e_fileCorrupt;
	}
	if (checkDataHash(dataHashType, &dataHash, &decrypted) != e_success) {
		printDebug("PwMDoc::openDoc(): checkDataHash() failed");
		f.close();
		return e_fileCorrupt;
	}
	if (!deSerializeDta(&decrypted, openLocked == 1)) {
		printDebug("PwMDoc::openDoc(): deSerializeDta() failed");
		f.close();
		return e_readFile;
	}
	f.close();
	startMpwTimer();
	setAutoLockTimeout(conf()->confGlobLockTimeout());
out_success:
	openDocList.edit(this, getTitle().latin1());
	emit docOpened(this);
	return e_success;
}

PwMerror PwMDoc::writeFileHeader(char keyHash, char dataHash, char crypt, char compress,
				 QString *pw, QFile *f)
{
	PWM_ASSERT(pw);
	PWM_ASSERT(f);
	PWM_ASSERT(listView);
	if (f->writeBlock(FILE_ID_HEADER, strlen(FILE_ID_HEADER)) !=
	    static_cast<Q_LONG>(strlen(FILE_ID_HEADER))) {
		return e_writeFile;
	}
	if (f->putch(PWM_FILE_VER) == -1 ||
	    f->putch(keyHash) == -1 ||
	    f->putch(dataHash) == -1 ||
	    f->putch(crypt) == -1 ||
	    f->putch(compress) == -1 ||
	    f->putch((getDocStatFlag(DOC_STAT_USE_CHIPCARD)) ?
		     (static_cast<char>(0x01)) : (static_cast<char>(0x00))) == -1) {
		return e_writeFile;
	}
	// write list-view size (width && height)
	uint32_t lvHeight;
	uint32_t lvWidth;
	if (listView->isMaximized()) {
		lvHeight = 0;
		lvWidth = 0;
	} else {
		QSize lvSize(listView->size());
		lvHeight = static_cast<uint32_t>(lvSize.height());
		lvWidth = static_cast<uint32_t>(lvSize.width());
	}
	char lvSizeBuf[8];
#ifdef BIG_ENDIAN_HOST
	lvSizeBuf[0] = (reinterpret_cast<char *>(&lvWidth))[0];
	lvSizeBuf[1] = (reinterpret_cast<char *>(&lvWidth))[1];
	lvSizeBuf[2] = (reinterpret_cast<char *>(&lvWidth))[2];
	lvSizeBuf[3] = (reinterpret_cast<char *>(&lvWidth))[3];
	lvSizeBuf[4] = (reinterpret_cast<char *>(&lvHeight))[0];
	lvSizeBuf[5] = (reinterpret_cast<char *>(&lvHeight))[1];
	lvSizeBuf[6] = (reinterpret_cast<char *>(&lvHeight))[2];
	lvSizeBuf[7] = (reinterpret_cast<char *>(&lvHeight))[3];
#else
	lvSizeBuf[0] = (reinterpret_cast<char *>(&lvWidth))[3];
	lvSizeBuf[1] = (reinterpret_cast<char *>(&lvWidth))[2];
	lvSizeBuf[2] = (reinterpret_cast<char *>(&lvWidth))[1];
	lvSizeBuf[3] = (reinterpret_cast<char *>(&lvWidth))[0];
	lvSizeBuf[4] = (reinterpret_cast<char *>(&lvHeight))[3];
	lvSizeBuf[5] = (reinterpret_cast<char *>(&lvHeight))[2];
	lvSizeBuf[6] = (reinterpret_cast<char *>(&lvHeight))[1];
	lvSizeBuf[7] = (reinterpret_cast<char *>(&lvHeight))[0];
#endif
	if (f->writeBlock(lvSizeBuf, sizeof(lvSizeBuf)) != sizeof(lvSizeBuf))
		return e_writeFile;

	// write bytes of NULL-data. These bytes are reserved for future-use.
	const int bufSize = 56;
	char tmp_buf[bufSize];
	memset(tmp_buf, 0x00, bufSize);
	if (f->writeBlock(tmp_buf, bufSize) != bufSize)
		return e_writeFile;

	if (keyHash == PWM_HASH_SHA1) {
		const int hashlen = SHA1_HASH_LEN_BYTE;
		Sha1 hash;
		hash.sha1_write(reinterpret_cast<const byte *>(pw->latin1()), pw->length());
		string ret = hash.sha1_read();
		if (f->writeBlock(ret.c_str(), hashlen) != hashlen)
			return e_writeFile;
	} else
		return e_hashNotImpl;
	return e_success;
}

PwMerror PwMDoc::checkHeader(char *cryptAlgo, QString *pw, char *compress,
			     unsigned int *headerLength, char *dataHashType,
			     string *dataHash, QFile *f)
{
	PWM_ASSERT(cryptAlgo);
	PWM_ASSERT(pw);
	PWM_ASSERT(headerLength);
	PWM_ASSERT(dataHashType);
	PWM_ASSERT(dataHash);
	PWM_ASSERT(f);
	int tmpRet;
	// check "magic" header
	const char magicHdr[] = FILE_ID_HEADER;
	const int hdrLen = array_size(magicHdr) - 1;
	char tmp[hdrLen];
	if (f->readBlock(tmp, hdrLen) != hdrLen)
		return e_readFile;
	if (memcmp(tmp, magicHdr, hdrLen) != 0)
		return e_fileFormat;
	// read and check file ver
	int fileV = f->getch();
	if (fileV == -1)
		return e_fileFormat;
	if (fileV != PWM_FILE_VER)
		return e_fileVer;
	// read hash hash type
	int keyHash = f->getch();
	if (keyHash == -1)
		return e_fileFormat;
	// read data hash type
	tmpRet = f->getch();
	if (tmpRet == -1)
		return e_fileFormat;
	*dataHashType = tmpRet;
	// read crypt algo
	tmpRet = f->getch();
	if (tmpRet == -1)
		return e_fileFormat;
	*cryptAlgo = tmpRet;
	// get compression-algo
	tmpRet = f->getch();
	if (tmpRet == -1)
		return e_fileFormat;
	*compress = tmpRet;
	// get the MPW-flag
	int mpw_flag = f->getch();
	if (mpw_flag == -1)
		return e_fileFormat;
	if (mpw_flag == 0x01)
		setDocStatFlag(DOC_STAT_USE_CHIPCARD);
	else
		unsetDocStatFlag(DOC_STAT_USE_CHIPCARD);
	// read window-size information
	QSize lvSize;
	uint32_t lvHeight, lvWidth;
	char lvSizeBuf[8];
	if (f->readBlock(lvSizeBuf, sizeof(lvSizeBuf)) != sizeof(lvSizeBuf))
		return e_readFile;
#ifdef BIG_ENDIAN_HOST
	(reinterpret_cast<char *>(&lvWidth))[0] = lvSizeBuf[0];
	(reinterpret_cast<char *>(&lvWidth))[1] = lvSizeBuf[1];
	(reinterpret_cast<char *>(&lvWidth))[2] = lvSizeBuf[2];
	(reinterpret_cast<char *>(&lvWidth))[3] = lvSizeBuf[3];
	(reinterpret_cast<char *>(&lvHeight))[0] = lvSizeBuf[4];
	(reinterpret_cast<char *>(&lvHeight))[1] = lvSizeBuf[5];
	(reinterpret_cast<char *>(&lvHeight))[2] = lvSizeBuf[6];
	(reinterpret_cast<char *>(&lvHeight))[3] = lvSizeBuf[7];
#else
	(reinterpret_cast<char *>(&lvWidth))[0] = lvSizeBuf[3];
	(reinterpret_cast<char *>(&lvWidth))[1] = lvSizeBuf[2];
	(reinterpret_cast<char *>(&lvWidth))[2] = lvSizeBuf[1];
	(reinterpret_cast<char *>(&lvWidth))[3] = lvSizeBuf[0];
	(reinterpret_cast<char *>(&lvHeight))[0] = lvSizeBuf[7];
	(reinterpret_cast<char *>(&lvHeight))[1] = lvSizeBuf[6];
	(reinterpret_cast<char *>(&lvHeight))[2] = lvSizeBuf[5];
	(reinterpret_cast<char *>(&lvHeight))[3] = lvSizeBuf[4];
#endif
	if (lvHeight == 0 || lvWidth == 0) {
		listView->hide();
		listView->showMaximized();
	} else {
		lvSize.setHeight(static_cast<int>(lvHeight));
		lvSize.setWidth(static_cast<int>(lvWidth));
		listView->resize(lvSize);
	}
	// skip 56-bytes. These are the "RESERVED"-bytes
	if (!(f->at(f->at() + 56)))
		return e_fileFormat;

	*pw = requestMpw(getDocStatFlag(DOC_STAT_USE_CHIPCARD));
	if (*pw == "") {
		/* the user didn't give a master-password
		 * or didn't insert a chipcard
		 */
		return e_noPw;
	}
	// verify key-hash
	switch (keyHash) {
		case PWM_HASH_SHA1: {
			// read hash from header
			const int hashLen = SHA1_HASH_LEN_BYTE;
			string readHash;
			int i;
			for (i = 0; i < hashLen; ++i)
				readHash.push_back(f->getch());
			Sha1 hash;
			hash.sha1_write(reinterpret_cast<const byte *>(pw->latin1()), pw->length());
			string ret = hash.sha1_read();
			if (ret != readHash)
				return e_wrongPw;	// hash doesn't match (wrong key)
			break;
		} default: {
			return e_hashNotImpl;
		}
	}
	// read the data-hash from the file
	int hashLen, i;
	switch (*dataHashType) {
		case PWM_HASH_SHA1: {
			hashLen = SHA1_HASH_LEN_BYTE;
			break;
		} default: {
			return e_hashNotImpl;
		}
	}
	*dataHash = "";
	for (i = 0; i < hashLen; ++i) {
		tmpRet = f->getch();
		if (tmpRet == -1)
			return e_fileFormat;
		dataHash->push_back(static_cast<char>(tmpRet));
	}
	*headerLength = f->at();
	return e_success;
}

PwMerror PwMDoc::writeDataHash(char dataHash, string *d, QFile *f)
{
	PWM_ASSERT(d);
	PWM_ASSERT(f);

	switch (dataHash) {
		case PWM_HASH_SHA1: {
			const int hashLen = SHA1_HASH_LEN_BYTE;
			Sha1 h;
			h.sha1_write(reinterpret_cast<const byte *>(d->c_str()), d->size());
			string hRet = h.sha1_read();
			if (f->writeBlock(hRet.c_str(), hashLen) != hashLen)
			    	return e_writeFile;
			break;
		} default: {
			return e_hashNotImpl;
		}
	}

	return e_success;
}

bool PwMDoc::backupFile(const QString &filePath)
{
	QFileInfo fi(filePath);
	if (!fi.exists())
		return true;
	QString pathOnly(fi.dirPath(true));
	QString nameOnly(fi.fileName());
	QString backupPath = pathOnly
			   + "/~"
			   + nameOnly
			   + ".backup";
	if (QFile::exists(backupPath))
		QFile::remove(backupPath);
	QFile origFile(filePath);
	if (!origFile.open(IO_ReadOnly))
		return false;
	QFile backupFile(backupPath);
	if (!backupFile.open(IO_ReadWrite)) {
		origFile.close();
		return false;
	}
	const int tmpBuf_size = 512;
	char tmpBuf[tmpBuf_size];
	Q_LONG bytesRead, bytesWritten;
	while (!origFile.atEnd()) {
		bytesRead = origFile.readBlock(tmpBuf,
				static_cast<Q_ULONG>(tmpBuf_size));
		if (bytesRead == -1) {
			origFile.close();
			backupFile.close();
			return false;
		}
		bytesWritten = backupFile.writeBlock(tmpBuf,
				static_cast<Q_ULONG>(bytesRead));
		if (bytesWritten != bytesRead) {
			origFile.close();
			backupFile.close();
			return false;
		}
	}
	origFile.close();
	backupFile.close();
	return true;
}

PwMerror PwMDoc::addEntry(const QString &category, PwMDataItem *d,
			  bool dontFlagDirty)
{
	PWM_ASSERT(d);
	unsigned int cat = 0;

	if (isDeepLocked()) {
		PwMerror ret;
		ret = deepLock(false);
		if (ret != e_success)
			return e_lock;
	}

	addCategory(category, &cat);

	if(numEntries(category) >= maxEntries)
		return e_maxAllowedEntr;

	vector<unsigned int> foundPositions;
	const int searchIn = SEARCH_IN_DESC | SEARCH_IN_NAME |
			     SEARCH_IN_URL | SEARCH_IN_LAUNCHER;
	findEntry(cat, *d, searchIn, &foundPositions, true);
	if (foundPositions.size()) {
		// DOH! We found this entry.
		return e_entryExists;
	}

	d->listViewPos = -1;
	d->lockStat = true;
	dta[cat].d.push_back(*d);

	delAllEmptyCat(true);

	if (!dontFlagDirty)
		flagDirty();
	return e_success;
}

PwMerror PwMDoc::addCategory(const QString &category, unsigned int *categoryIndex,
			     bool checkIfExist)
{
	if (isDeepLocked()) {
		PwMerror ret;
		ret = deepLock(false);
		if (ret != e_success)
			return e_lock;
	}
	if (checkIfExist) {
		if (findCategory(category, categoryIndex))
			return e_categoryExists;
	}
	PwMCategoryItem item;
	item.name = category.latin1();
	dta.push_back(item);
	if (categoryIndex)
		*categoryIndex = dta.size() - 1;
	return e_success;
}

bool PwMDoc::delEntry(const QString &category, unsigned int index, bool dontFlagDirty)
{
	unsigned int cat = 0;

	if (!findCategory(category, &cat)) {
		BUG();
		return false;
	}

	return delEntry(cat, index, dontFlagDirty);
}

bool PwMDoc::delEntry(unsigned int category, unsigned int index, bool dontFlagDirty)
{
	if (isDeepLocked())
		return false;
	if (index > dta[category].d.size() - 1)
		return false;
#ifndef CONFIG_DEL_WO_PW
	if (!lockAt(category, index, false, false))
		return false;
#endif
	int lvPos = dta[category].d[index].listViewPos;

	// delete entry
	dta[category].d.erase(dta[category].d.begin() + index);

	unsigned int i, entries = numEntries(category);
	if (!entries) {
		// no more entries in this category, so
		// we can delete it, too.
		if (!delCategory(category))
			BUG();
		// delCategory() flags it dirty, so we need not to do so.
		return true;
	}
	for (i = 0; i < entries; ++i) {
		// decrement all listViewPositions that are greater than the deleted.
		if (dta[category].d[i].listViewPos > lvPos)
			--dta[category].d[i].listViewPos;
	}

	if (!dontFlagDirty)
		flagDirty();
	return true;
}

bool PwMDoc::editEntry(const QString &oldCategory, const QString &newCategory,
		       unsigned int index, PwMDataItem * d)
{
	PWM_ASSERT(d);
	unsigned int oldCat = 0;

	if (!findCategory(oldCategory, &oldCat)) {
		BUG();
		return -1;
	}

	return editEntry(oldCat, newCategory, index, d);
}

bool PwMDoc::editEntry(unsigned int oldCategory, const QString &newCategory,
		       unsigned int index, PwMDataItem *d)
{
	if (isDeepLocked())
		return false;
	if (dta[oldCategory].name != newCategory.latin1()) {
		// the user changed the category.
		PwMerror ret;
		ret = addEntry(newCategory, d, true);
		if (ret != e_success)
			return false;
		if (!delEntry(oldCategory, index, true))
			return false;
	} else {
		dta[oldCategory].d[index] = *d;
	}

	flagDirty();
	return true;
}

unsigned int PwMDoc::numEntries(const QString &category)
{
	unsigned int cat = 0;

	if (!findCategory(category, &cat)) {
		BUG();
		return 0;
	}

	return numEntries(cat);
}

bool PwMDoc::serializeDta(string *d)
{
	PWM_ASSERT(d);
	Serializer ser;
	if (!ser.serialize(dta))
		return false;
	d->assign(ser.getXml());
	if (!d->size())
		return false;
	return true;
}

bool PwMDoc::deSerializeDta(const string *d, bool entriesLocked)
{
	PWM_ASSERT(d);
	try {
		Serializer ser(d->c_str());
		ser.setDefaultLockStat(entriesLocked);
		if (!ser.deSerialize(&dta))
			return false;
	} catch (PwMException) {
		return false;
	}
	emit dataChanged(this);
	return true;
}

bool PwMDoc::getEntry(const QString &category, unsigned int index,
		      PwMDataItem * d, bool unlockIfLocked)
{
	PWM_ASSERT(d);
	unsigned int cat = 0;

	if (!findCategory(category, &cat)) {
		BUG();
		return false;
	}

	return getEntry(cat, index, d, unlockIfLocked);
}

bool PwMDoc::getEntry(unsigned int category, unsigned int index,
		      PwMDataItem *d, bool unlockIfLocked)
{
	if (index > dta[category].d.size() - 1)
		return false;

	bool locked = isLocked(category, index);
	if (locked) {
		/* this entry is locked. We don't return a password,
		 * until it's unlocked by the user by inserting
		 * chipcard or entering the mpw
		 */
		if (unlockIfLocked) {
			if (!lockAt(category, index, false)) {
				return false;
			}
			locked = false;
		}
	}

	*d = dta[category].d[index];
	if (locked)
		d->pw = LOCKED_STRING.latin1();

	return true;
}

void PwMDoc::getCommentByLvp(const QString &category, int listViewPos,
			     string *foundComment)
{
	PWM_ASSERT(foundComment);
	unsigned int cat = 0;

	if (!findCategory(category, &cat)) {
		BUG();
		return;
	}

	unsigned int i, entries = numEntries(cat);
	for (i = 0; i < entries; ++i) {
		if (dta[cat].d[i].listViewPos == listViewPos) {
			*foundComment = dta[cat].d[i].comment;
			return;
		}
	}
	BUG();
}

bool PwMDoc::compressDta(string *d, char algo)
{
	PWM_ASSERT(d);
	switch (algo) {
		case PWM_COMPRESS_GZIP: {
			CompressGzip comp;
			return comp.compress(d);
		} case PWM_COMPRESS_BZIP2: {
			CompressBzip2 comp;
			return comp.compress(d);
		} case PWM_COMPRESS_NONE: {
			return true;
		} default: {
			BUG();
		}
	}
	return false;
}

bool PwMDoc::decompressDta(string *d, char algo)
{
	PWM_ASSERT(d);
	switch (algo) {
		case PWM_COMPRESS_GZIP: {
			CompressGzip comp;
			return comp.decompress(d);
		} case PWM_COMPRESS_BZIP2: {
			CompressBzip2 comp;
			return comp.decompress(d);
		} case PWM_COMPRESS_NONE: {
			return true;
		}
	}
	return false;
}

PwMerror PwMDoc::encrypt(string *d, const QString *pw, QFile *f, char algo)
{
	PWM_ASSERT(d);
	PWM_ASSERT(pw);
	PWM_ASSERT(f);

	if (algo == PWM_CRYPT_BLOWFISH)
		Blowfish::padNull(d);
	unsigned int encSize = d->length();
	byte *encrypted = new byte[encSize];

	if (algo == PWM_CRYPT_BLOWFISH) {
		Blowfish bf;
		if (bf.bf_setkey((byte *) pw->latin1(), pw->length())) {
			delete [] encrypted;
			return e_weakPw;
		}
		bf.bf_encrypt((byte *) encrypted, (byte *) d->c_str(), encSize);
	} else {
		delete [] encrypted;
		return e_cryptNotImpl;
	}

	// write encrypted data to file
	if (f->writeBlock((char *) encrypted, encSize) != (int) encSize) {
		delete [] encrypted;
		return e_writeFile;
	}

	unsigned int i;
	for (i = 0; i < encSize; ++i)
		encrypted[i] = '\0';
	delete [] encrypted;

	return e_success;
}

PwMerror PwMDoc::decrypt(string *d, unsigned int pos, const QString *pw,
			 char algo, QFile *f)
{
	PWM_ASSERT(d);
	PWM_ASSERT(pw);
	PWM_ASSERT(f);

	unsigned int cryptLen = f->size() - pos;
	byte *encrypted = new byte[cryptLen];
	byte *decrypted = new byte[cryptLen];

	f->at(pos);
	if (f->readBlock((char *) encrypted, cryptLen) != (int) cryptLen) {
		delete [] encrypted;
		delete [] decrypted;
		return e_readFile;
	}

	if (algo == PWM_CRYPT_BLOWFISH) {
		Blowfish bf;
		bf.bf_setkey((byte *) pw->latin1(), pw->length());
		bf.bf_decrypt(decrypted, encrypted, cryptLen);
	} else {
		delete [] encrypted;
		delete [] decrypted;
		return e_cryptNotImpl;
	}
	unsigned int i;
	for (i = 0; i < cryptLen; ++i)
		encrypted[i] = '\0';
	delete [] encrypted;

	for (i = 0; i < cryptLen; ++i) {
		d->push_back(decrypted[i]);
		decrypted[i] = '\0';
	}
	delete [] decrypted;
	if (algo == PWM_CRYPT_BLOWFISH) {
		if (!Blowfish::unpadNull(d)) {
			BUG();
			return e_readFile;
		}
	}

	return e_success;
}

PwMerror PwMDoc::checkDataHash(char dataHashType, const string *dataHash,
			       const string *dataStream)
{
	PWM_ASSERT(dataHash);
	PWM_ASSERT(dataStream);
	if (dataHashType == PWM_HASH_SHA1) {
		Sha1 hash;
		hash.sha1_write((byte*)dataStream->c_str(), dataStream->length());
		string ret = hash.sha1_read();
		if (ret != *dataHash)
			return e_fileCorrupt;
	} else
		return e_hashNotImpl;
	return e_success;
}

bool PwMDoc::lockAt(unsigned int category, unsigned int index,
		    bool lock, bool emitDataChangedSig)
{
	if (index >= numEntries(category)) {
		BUG();
		return false;
	}
	if (lock == dta[category].d[index].lockStat)
		return true;

	if (!lock && currentPw != "") {
		// "unlocking" and "password is already set"
		if (!getDocStatFlag(DOC_STAT_UNLOCK_WITHOUT_PW)) {
			// unlocking without pw not allowed
			QString pw;
			pw = requestMpw(getDocStatFlag(DOC_STAT_USE_CHIPCARD));
			if (pw != "") {
				if (pw != currentPw) {
					wrongMpwMsgBox(getDocStatFlag(DOC_STAT_USE_CHIPCARD));
					return false;
				} else
					startMpwTimer();
			} else {
				return false;
			}
		} else
			startMpwTimer();
	}

	dta[category].d[index].lockStat = lock;

	if (emitDataChangedSig)
		emit dataChanged(this);
	if (!lock)
		startAutoLockTimer();

	return true;

}

bool PwMDoc::lockAt(const QString &category,unsigned int index,
		    bool lock,  bool emitDataChangedSig)
{
	unsigned int cat = 0;

	if (!findCategory(category, &cat)) {
		BUG();
		return false;
	}

	return lockAt(cat, index, lock, emitDataChangedSig);
}

bool PwMDoc::lockAll(bool lock)
{
	if (!lock && isDeepLocked()) {
		PwMerror ret;
		ret = deepLock(false);
		if (ret != e_success)
			return false;
		return true;
	}
	if (isDocEmpty()) {
		return true;
	}
	if (!lock && currentPw != "") {
		// unlocking and password is already set
		if (!getDocStatFlag(DOC_STAT_UNLOCK_WITHOUT_PW)) {
			// unlocking without pw not allowed
			QString pw;
			pw = requestMpw(getDocStatFlag(DOC_STAT_USE_CHIPCARD));
			if (pw != "") {
				if (pw != currentPw) {
					wrongMpwMsgBox(getDocStatFlag(DOC_STAT_USE_CHIPCARD));
					return false;
				} else
					startMpwTimer();
			} else {
				return false;
			}
		} else
			startMpwTimer();
	}

	vector<PwMCategoryItem>::iterator catBegin = dta.begin(),
					  catEnd = dta.end(),
					  catI = catBegin;
	vector<PwMDataItem>::iterator entrBegin, entrEnd, entrI;
	while (catI != catEnd) {
		entrBegin = catI->d.begin();
		entrEnd = catI->d.end();
		entrI = entrBegin;
		while (entrI != entrEnd) {
			entrI->lockStat = lock;
			++entrI;
		}
		++catI;
	}

	emit dataChanged(this);
	if (lock)
		stopAutoLockTimer();
	else
		startAutoLockTimer();

	return true;
}

bool PwMDoc::isLocked(const QString &category, unsigned int index)
{
	unsigned int cat = 0;

	if (!findCategory(category, &cat)) {
		BUG();
		return false;
	}

	return isLocked(cat, index);
}

void PwMDoc::autoLockDocument()
{
	if (conf()->confGlobAutoDeepLock()) {
		deepLock(true);
	} else {
		lockAll(true);
	}
}

void PwMDoc::setAutoLockTimeout(int _autoLockTimerValue,
				bool dontStartNewTimer)
{
	stopAutoLockTimer();
	autoLockTimerValue = _autoLockTimerValue;
	if (!dontStartNewTimer && autoLockTimerValue > 0)
		startAutoLockTimer();
}

void PwMDoc::startAutoLockTimer()
{
	if (autoLockTimer->isActive())
		autoLockTimer->stop();
	if (autoLockTimerValue > 0)
		autoLockTimer->start(autoLockTimerValue * 1000, TRUE);
}

bool PwMDoc::unlockAll_tempoary(bool revert)
{
	static vector< vector<bool> > *oldLockStates = 0;
	static bool wasDeepLocked;

	if (revert) {	// revert the unlocking
		if (oldLockStates) {
			/* we actually _have_ unlocked something, because
			 * we have allocated space for the oldLockStates.
			 * So, go on and revert them!
			 */
			if (wasDeepLocked) {
				PwMerror ret = deepLock(true);
				if (ret == e_success) {
					/* deep-lock succeed. We are save.
					 * (but if it failed, just go on
					 * lock them normally)
					 */
					delete_and_null(oldLockStates);
					setAutoLockTimeout(conf()->confGlobLockTimeout());
					printDebug("tempoary unlocking of dta "
						   "reverted by deep-locking.");
					return true;
				}
				printDebug("deep-lock failed while reverting! "
					   "Falling back to normal-lock.");
			}
			if (unlikely(!wasDeepLocked &&
				     numCategories() != oldLockStates->size())) {
				/* DOH! We have modified "dta" while
				 * it was unlocked tempoary. DON'T DO THIS!
				 */
				BUG();
				delete_and_null(oldLockStates);
				setAutoLockTimeout(conf()->confGlobLockTimeout());
				return false;
			}
			vector<PwMCategoryItem>::iterator catBegin = dta.begin(),
							  catEnd = dta.end(),
							  catI = catBegin;
			vector<PwMDataItem>::iterator entrBegin, entrEnd, entrI;
			vector< vector<bool> >::iterator oldCatStatI = oldLockStates->begin();
			vector<bool>::iterator oldEntrStatBegin,
					       oldEntrStatEnd,
					       oldEntrStatI;
			while (catI != catEnd) {
				entrBegin = catI->d.begin();
				entrEnd = catI->d.end();
				entrI = entrBegin;
				if (likely(!wasDeepLocked)) {
					oldEntrStatBegin = oldCatStatI->begin();
					oldEntrStatEnd = oldCatStatI->end();
					oldEntrStatI = oldEntrStatBegin;
					if (unlikely(catI->d.size() != oldCatStatI->size())) {
						/* DOH! We have modified "dta" while
						 * it was unlocked tempoary. DON'T DO THIS!
						 */
						BUG();
						delete_and_null(oldLockStates);
						setAutoLockTimeout(conf()->confGlobLockTimeout());
						return false;
					}
				}
				while (entrI != entrEnd) {
					if (wasDeepLocked) {
						/* this is an error-fallback if
						 * deeplock didn't succeed
						 */
						entrI->lockStat = true;
					} else {
						entrI->lockStat = *oldEntrStatI;
					}
					++entrI;
					if (likely(!wasDeepLocked))
						++oldEntrStatI;
				}
				++catI;
				if (likely(!wasDeepLocked))
					++oldCatStatI;
			}
			delete_and_null(oldLockStates);
			if (unlikely(wasDeepLocked)) {
				/* error fallback... */
				unsetDocStatFlag(DOC_STAT_DEEPLOCKED);
				emit dataChanged(this);
				printDebug("WARNING: unlockAll_tempoary(true) "
					   "deeplock fallback!");
			}
			printDebug("tempoary unlocking of dta reverted.");
		} else {
			printDebug("unlockAll_tempoary(true): nothing to do.");
		}
		setAutoLockTimeout(conf()->confGlobLockTimeout());
	} else {	// unlock all data tempoary
		if (unlikely(oldLockStates != 0)) {
			/* DOH! We have already unlocked the data tempoarly.
			 * No need to do it twice. ;)
			 */
			BUG();
			return false;
		}
		wasDeepLocked = false;
		bool mustUnlock = false;
		if (isDeepLocked()) {
			PwMerror ret;
			while (1) {
				ret = deepLock(false);
				if (ret == e_success) {
					break;
				} else if (ret == e_wrongPw) {
					wrongMpwMsgBox(getDocStatFlag(DOC_STAT_USE_CHIPCARD));
				} else {
					printDebug("deep-unlocking failed while "
						   "tempoary unlocking!");
					return false;
				}
			}
			wasDeepLocked = true;
			mustUnlock = true;
		} else {
			// first check if it's needed to unlock some entries
			vector<PwMCategoryItem>::iterator catBegin = dta.begin(),
							  catEnd = dta.end(),
							  catI = catBegin;
			vector<PwMDataItem>::iterator entrBegin, entrEnd, entrI;
			while (catI != catEnd) {
				entrBegin = catI->d.begin();
				entrEnd = catI->d.end();
				entrI = entrBegin;
				while (entrI != entrEnd) {
					if (entrI->lockStat == true) {
						mustUnlock = true;
						break;
					}
					++entrI;
				}
				if (mustUnlock)
					break;
				++catI;
			}
		}
		if (!mustUnlock) {
			// nothing to do.
			stopAutoLockTimer();
			printDebug("unlockAll_tempoary(): nothing to do.");
			return true;
		} else if (!wasDeepLocked) {
			if (!getDocStatFlag(DOC_STAT_UNLOCK_WITHOUT_PW) &&
			    currentPw != "") {
				/* we can't unlock without mpw, so
				 * we need to ask for it.
				 */
				QString pw;
				while (1) {
					pw = requestMpw(getDocStatFlag(DOC_STAT_USE_CHIPCARD));
					if (pw == "") {
						return false;
					} else if (pw == currentPw) {
						break;
					}
					wrongMpwMsgBox(getDocStatFlag(DOC_STAT_USE_CHIPCARD));
				}
			}
		}
		stopAutoLockTimer();
		oldLockStates = new vector< vector<bool> >;
		vector<bool> tmp_vec;
		vector<PwMCategoryItem>::iterator catBegin = dta.begin(),
						  catEnd = dta.end(),
						  catI = catBegin;
		vector<PwMDataItem>::iterator entrBegin, entrEnd, entrI;
		while (catI != catEnd) {
			entrBegin = catI->d.begin();
			entrEnd = catI->d.end();
			entrI = entrBegin;
			while (entrI != entrEnd) {
				if (!wasDeepLocked) {
					tmp_vec.push_back(entrI->lockStat);
				}
				entrI->lockStat = false;
				++entrI;
			}
			if (!wasDeepLocked) {
				oldLockStates->push_back(tmp_vec);
				tmp_vec.clear();
			}
			++catI;
		}
		printDebug("tempoary unlocked dta.");
	}

	return true;
}

PwMerror PwMDoc::deepLock(bool lock, bool saveToFile)
{
	PwMerror ret;

	if (lock) {
		if (isDeepLocked())
			return e_lock;
		if (saveToFile) {
			if (isDocEmpty())
				return e_docIsEmpty;
			ret = saveDoc(conf()->confGlobCompression());
			if (ret == e_filename) {
				/* the doc wasn't saved to a file
				 * by the user, yet.
				 */
				cantDeeplock_notSavedMsgBox();
				return e_docNotSaved;
			} else if (ret != e_success) {
				return e_lock;
			}
		}
		stopAutoLockTimer();
		clearDoc();
		PwMDataItem d;
		d.desc = IS_DEEPLOCKED_SHORTMSG.latin1();
		d.comment = IS_DEEPLOCKED_MSG.latin1();
		d.listViewPos = 0;
		addEntry(DEFAULT_CATEGORY, &d, true);
		lockAt(DEFAULT_CATEGORY, 0, true);
		unsetDocStatFlag(DOC_STAT_DISK_DIRTY);
		setDocStatFlag(DOC_STAT_DEEPLOCKED);
	} else {
		if (!isDeepLocked())
			return e_lock;
		ret = openDoc(&filename, (conf()->confGlobUnlockOnOpen())
					  ? 0 : 1);
		if (ret == e_wrongPw) {
			return e_wrongPw;
		} else if (ret != e_success) {
			printDebug(string("PwMDoc::deepLock(false): ERR! openDoc() == ")
				   + tostr(static_cast<int>(ret)));
			return e_lock;
		}
		unsetDocStatFlag(DOC_STAT_DEEPLOCKED);
		setAutoLockTimeout(conf()->confGlobLockTimeout());
	}

	emit dataChanged(this);
	return e_success;
}

void PwMDoc::_deepUnlock()
{
	deepLock(false);
}

void PwMDoc::clearDoc()
{
	dta.clear();
	PwMCategoryItem d;
	d.name = DEFAULT_CATEGORY.latin1();
	dta.push_back(d);
	currentPw = "";
	unsetDocStatFlag(DOC_STAT_UNLOCK_WITHOUT_PW);
}

void PwMDoc::resetUnlockWoPw()
{
	unsetDocStatFlag(DOC_STAT_UNLOCK_WITHOUT_PW);
}

void PwMDoc::startMpwTimer()
{
	if (mpwTimer->isActive())
		mpwTimer->stop();
	setDocStatFlag(DOC_STAT_UNLOCK_WITHOUT_PW);
	mpwTimer->start(mpwTimerValue * 1000, TRUE);
}

void PwMDoc::changeCurrentPw()
{
	if (currentPw == "")
		return; // doc hasn't been saved. No mpw available.
	bool useChipcard = getDocStatFlag(DOC_STAT_USE_CHIPCARD);
	QString pw = requestMpwChange(&currentPw, &useChipcard);
	if (pw == "")
		return;
	if (useChipcard)
		setDocStatFlag(DOC_STAT_USE_CHIPCARD);
	else
		unsetDocStatFlag(DOC_STAT_USE_CHIPCARD);
	setCurrentPw(pw);
}

void PwMDoc::setListViewPos(const QString &category, unsigned int index,
			    int pos)
{
	unsigned int cat = 0;

	if (!findCategory(category, &cat)) {
		BUG();
		return;
	}

	dta[cat].d[index].listViewPos = pos;

/* FIXME workaround: don't flag dirty, because this function sometimes
 * get's called when it shouldn't. It's because PwMView assumes
 * the user resorted the UI on behalf of signal layoutChanged().
 * This is somewhat broken and incorrect, but I've no other
 * solution for now.
 */
//	setDocStatFlag(DOC_STAT_DISK_DIRTY);
}

int PwMDoc::getListViewPos(const QString &category, unsigned int index)
{
	unsigned int cat = 0;

	if (!findCategory(category, &cat)) {
		BUG();
		return -1;
	}

	return dta[cat].d[index].listViewPos;
}

void PwMDoc::findEntry(unsigned int category, PwMDataItem find, unsigned int searchIn,
		       vector<unsigned int> *foundPositions, bool breakAfterFound,
		       bool caseSensitive, bool exactWordMatch, bool sortByLvp)
{
	PWM_ASSERT(foundPositions);
	PWM_ASSERT(searchIn);
	foundPositions->clear();

	unsigned int i, entries = numEntries(category);
	for (i = 0; i < entries; ++i) {
		if (searchIn & SEARCH_IN_DESC) {
			if (!compareString(find.desc, dta[category].d[i].desc,
					  caseSensitive, exactWordMatch)) {
				continue;
			}
		}
		if (searchIn & SEARCH_IN_NAME) {
			if (!compareString(find.name, dta[category].d[i].name,
					  caseSensitive, exactWordMatch)) {
				continue;
			}
		}
		if (searchIn & SEARCH_IN_PW) {
			bool wasLocked = isLocked(category, i);
			lockAt(category, i, false, false);
			if (!compareString(find.pw, dta[category].d[i].pw,
					  caseSensitive, exactWordMatch)) {
				continue;
			}
			lockAt(category, i, wasLocked, false);
		}
		if (searchIn & SEARCH_IN_COMMENT) {
			if (!compareString(find.comment, dta[category].d[i].comment,
					  caseSensitive, exactWordMatch)) {
				continue;
			}
		}
		if (searchIn & SEARCH_IN_URL) {
			if (!compareString(find.url, dta[category].d[i].url,
					  caseSensitive, exactWordMatch)) {
				continue;
			}
		}
		if (searchIn & SEARCH_IN_LAUNCHER) {
			if (!compareString(find.launcher, dta[category].d[i].launcher,
					  caseSensitive, exactWordMatch)) {
				continue;
			}
		}

		// all selected "searchIn" matched.
		foundPositions->push_back(i);
		if (breakAfterFound)
			break;
	}

	if (sortByLvp && foundPositions->size() > 1) {
		vector< pair<unsigned int /* foundPosition (real doc pos) */,
			     unsigned int /* lvp-pos */> > tmp_vec;
		
		unsigned int i, items = foundPositions->size();
		pair<unsigned int, unsigned int> tmp_pair;
		for (i = 0; i < items; ++i) {
			tmp_pair.first = (*foundPositions)[i];
			tmp_pair.second = dta[category].d[(*foundPositions)[i]].listViewPos;
			tmp_vec.push_back(tmp_pair);
		}
		sort(tmp_vec.begin(), tmp_vec.end(), dta_lvp_greater());
		foundPositions->clear();
		for (i = 0; i < items; ++i) {
			foundPositions->push_back(tmp_vec[i].first);
		}
	}
}

void PwMDoc::findEntry(const QString &category, PwMDataItem find, unsigned int searchIn,
		       vector<unsigned int> *foundPositions, bool breakAfterFound,
		       bool caseSensitive, bool exactWordMatch, bool sortByLvp)
{
	PWM_ASSERT(foundPositions);
	unsigned int cat = 0;

	if (!findCategory(category, &cat)) {
		foundPositions->clear();
		return;
	}

	findEntry(cat, find, searchIn, foundPositions, breakAfterFound,
		  caseSensitive, exactWordMatch, sortByLvp);
}

bool PwMDoc::compareString(const string &s1, const string &s2, bool caseSensitive,
			   bool exactWordMatch)
{
	QString _s1(s1.c_str());
	QString _s2(s2.c_str());
	if (!caseSensitive) {
		_s1 = _s1.lower();
		_s2 = _s2.lower();
	}
	if (exactWordMatch ? (_s1 == _s2) : (_s2.find(_s1) != -1))
		return true;
	return false;
}

bool PwMDoc::findCategory(const QString &name, unsigned int *index)
{
	unsigned int i, cat = numCategories();
	if (!cat)
		return false;

	for (i = 0; i < cat; ++i) {
		if (dta[i].name == name.latin1()) {
			if (index)
				*index = i;
			return true;
		}
	}

	return false;
}

bool PwMDoc::renameCategory(const QString &category, const QString &newName)
{
	unsigned int cat = 0;

	if (!findCategory(category, &cat))
		return false;

	return renameCategory(cat, newName);
}

bool PwMDoc::renameCategory(unsigned int category, const QString &newName,
			    bool dontFlagDirty)
{
	if (category > numCategories() - 1)
		return false;

	dta[category].name = newName.latin1();
	if (!dontFlagDirty)
		flagDirty();

	return true;
}

bool PwMDoc::delCategory(const QString &category)
{
	unsigned int cat = 0;

	if (!findCategory(category, &cat))
		return false;

	return delCategory(cat);
}

bool PwMDoc::delCategory(unsigned int category, bool dontFlagDirty)
{
	if (category > numCategories() - 1)
		return false;

	// We don't delete it, if it is the last existing
	// category! Instead we rename it to "Default".
	if (numCategories() > 1) {
		dta.erase(dta.begin() + category);
	} else {
		renameCategory(category, DEFAULT_CATEGORY, dontFlagDirty);
		return true;
	}
	if (!dontFlagDirty)
		flagDirty();

	return true;
}

void PwMDoc::delAllEmptyCat(bool dontFlagDirty)
{
	vector<PwMCategoryItem>::iterator begin = dta.begin(),
					  end = dta.end(),
					  i = begin;
	while (i != end) {
		if (i->d.empty()) {
			delCategory(begin - i, dontFlagDirty);
		}
		++i;
	}
}

void PwMDoc::getCategoryList(vector<string> *list)
{
	PWM_ASSERT(list);
	list->clear();
	vector<PwMCategoryItem>::iterator i = dta.begin(),
					  end = dta.end();
	while (i != end) {
		list->push_back(i->name);
		++i;
	}
}

void PwMDoc::getCategoryList(QStringList *list)
{
	PWM_ASSERT(list);
	list->clear();
	vector<PwMCategoryItem>::iterator i = dta.begin(),
					  end = dta.end();
	while (i != end) {
		list->push_back(i->name.c_str());
		++i;
	}
}

void PwMDoc::getEntryList(const QString &category, QStringList *list)
{
	PWM_ASSERT(list);
	unsigned int cat = 0;
	if (!findCategory(category, &cat)) {
		list->clear();
		return;
	}
	getEntryList(cat, list);
}

void PwMDoc::getEntryList(const QString &category, vector<string> *list)
{
	PWM_ASSERT(list);
	unsigned int cat = 0;
	if (!findCategory(category, &cat)) {
		list->clear();
		return;
	}
	getEntryList(cat, list);
}

void PwMDoc::getEntryList(unsigned int category, vector<string> *list)
{
	PWM_ASSERT(list);
	list->clear();
	vector<PwMDataItem>::iterator begin = dta[category].d.begin(),
				      end = dta[category].d.end(),
				      i = begin;
	while (i != end) {
		list->push_back(i->desc);
		++i;
	}
}

void PwMDoc::getEntryList(unsigned int category, QStringList *list)
{
	PWM_ASSERT(list);
	list->clear();
	vector<PwMDataItem>::iterator begin = dta[category].d.begin(),
				      end = dta[category].d.end(),
				      i = begin;
	while (i != end) {
		list->push_back(i->desc.c_str());
		++i;
	}
}

void PwMDoc::setMpwTimeout(int _mpwTimerValue, bool dontStartNewTimer)
{
	if (dontStartNewTimer) {
		// kill old timer
		mpwTimerValue = 0;
		startMpwTimer();
	}
	mpwTimerValue = _mpwTimerValue;
	if (!dontStartNewTimer)
		startMpwTimer();
}

bool PwMDoc::execLauncher(const QString &category, unsigned int entryIndex)
{
	unsigned int cat = 0;

	if (!findCategory(category, &cat))
		return false;

	return execLauncher(cat, entryIndex);
}

bool PwMDoc::execLauncher(unsigned int category, unsigned int entryIndex)
{
	if (geteuid() == 0) {
		rootAlertMsgBox();
		return false;
	}
	QString command(dta[category].d[entryIndex].launcher.c_str());
	bool wasLocked = isLocked(category, entryIndex);

	if (command.find("$p") != -1) {
		/* the user requested the password to be included
		 * into the command. We have to ask for the password,
		 * if it's locked. We do that by unlocking the entry
		 */
		if (!lockAt(category, entryIndex, false))
			return false;
	}

	command.replace("$d", dta[category].d[entryIndex].desc.c_str());
	command.replace("$n", dta[category].d[entryIndex].name.c_str());
	command.replace("$p", dta[category].d[entryIndex].pw.c_str());
	command.replace("$u", dta[category].d[entryIndex].url.c_str());
	command.replace("$c", dta[category].d[entryIndex].comment.c_str());
	command.append(" &");

	QString customXterm(conf()->confGlobXtermCommand());
	if (!customXterm.isEmpty())
		command = customXterm + " " + command;

	system(command.latin1());

	lockAt(category, entryIndex, wasLocked);
	return true;
}

bool PwMDoc::goToURL(const QString &category, unsigned int entryIndex)
{
	unsigned int cat = 0;

	if (!findCategory(category, &cat))
		return false;

	return goToURL(cat, entryIndex);
}

bool PwMDoc::goToURL(unsigned int category, unsigned int entryIndex)
{
	if (geteuid() == 0) {
		rootAlertMsgBox();
		return false;
	}
	QString url(dta[category].d[entryIndex].url.c_str());
	if (url.isEmpty())
		return false;

	QString customBrowser(conf()->confGlobBrowserCommand());
	if (!customBrowser.isEmpty()) {
		browserProc.clearArguments();
		browserProc << customBrowser << url;
		if (browserProc.start(KProcess::DontCare))
			return true;
	}

	browserProc.clearArguments();
	browserProc << "konqueror" << url;
	if (browserProc.start(KProcess::DontCare))
		return true;

	browserProc.clearArguments();
	browserProc << "mozilla" << url;
	if (browserProc.start(KProcess::DontCare))
		return true;

	browserProc.clearArguments();
	browserProc << "opera" << url;
	if (browserProc.start(KProcess::DontCare))
		return true;
	return false;
}

PwMerror PwMDoc::exportToText(const QString *file)
{
	PWM_ASSERT(file);
	if (QFile::exists(*file)) {
		if (!QFile::remove(*file))
			return e_accessFile;
	}
	QFile f(*file);
	if (!f.open(IO_ReadWrite))
		return e_openFile;

	if (!unlockAll_tempoary()) {
		f.close();
		return e_lock;
	}

	// write header
	string header = i18n("Password table generated by\nPwM v").latin1();
	header += VERSION;
	header += i18n("\non ").latin1();
	QDate currDate = QDate::currentDate();
	header += currDate.toString("ddd MMMM d ").latin1();
	QTime currTime = QTime::currentTime();
	header += currTime.toString("hh:mm:ss ").latin1();
	header += tostr(currDate.year());
	header += "\n==============================\n\n";

	if (f.writeBlock(header.c_str(), header.length()) != (Q_LONG)header.length()) {
		unlockAll_tempoary(true);
		f.close();
		return e_writeFile;
	}

	unsigned int i, numCat = numCategories();
	unsigned int j, numEnt;
	string exp;
	for (i = 0; i < numCat; ++i) {
		numEnt = numEntries(i);

		exp = "\n== Category: ";
		exp += dta[i].name;
		exp += " ==\n";
		if (f.writeBlock(exp.c_str(), exp.length()) != (Q_LONG)exp.length()) {
				unlockAll_tempoary(true);
				f.close();
				return e_writeFile;
		}

		for (j = 0; j < numEnt; ++j) {
			exp = "\n-- ";
			exp += dta[i].d[j].desc;
			exp += " --\n";

			exp += i18n("Username: ").latin1();
			exp += dta[i].d[j].name;
			exp += "\n";

			exp += i18n("Password: ").latin1();
			exp += dta[i].d[j].pw;
			exp += "\n";

			exp += i18n("Comment: ").latin1();
			exp += dta[i].d[j].comment;
			exp += "\n";

			exp += i18n("URL: ").latin1();
			exp += dta[i].d[j].url;
			exp += "\n";

			exp += i18n("Launcher: ").latin1();
			exp += dta[i].d[j].launcher;
			exp += "\n";

			if (f.writeBlock(exp.c_str(), exp.length()) != (Q_LONG)exp.length()) {
				unlockAll_tempoary(true);
				f.close();
				return e_writeFile;
			}
		}
	}
	unlockAll_tempoary(true);
	f.close();

	return e_success;
}

PwMerror PwMDoc::importFromText(const QString *file, int format)
{
	PWM_ASSERT(file);
	if (format == 0)
		return importText_PwM(file);
	else if (format == -1) {
		// probe for all formats
		if (importText_PwM(file) == e_success)
			return e_success;
		dta.clear();
		emit dataChanged(this);
		// add next format here...
		return e_fileFormat;
	}
	return e_invalidArg;
}

PwMerror PwMDoc::importText_PwM(const QString *file)
{
	PWM_ASSERT(file);
	FILE *f;
	int tmp;
	ssize_t ret;
	string curCat;
	unsigned int entriesRead = 0;
	PwMDataItem currItem;
	f = fopen(file->latin1(), "r");
	if (!f)
		return e_openFile;
	size_t ch_tmp_size = 1024;
	char *ch_tmp = (char*)malloc(ch_tmp_size);
	if (!ch_tmp) {
		fclose(f);
		return e_outOfMem;
	}

	// - check header
	if (getline(&ch_tmp, &ch_tmp_size, f) == -1) // skip first line.
		goto formatError;
	// check version-string and return version in "ch_tmp".
	if (fscanf(f, "PwM v%s", ch_tmp) != 1) {
		// header not recognized as PwM generated header
		goto formatError;
	}
	// set filepointer behind version-string-line previously checked
	if (getline(&ch_tmp, &ch_tmp_size, f) == -1)
		goto formatError;
	// skip next line containing the build-date
	if (getline(&ch_tmp, &ch_tmp_size, f) == -1)
		goto formatError;
	// read header termination line
	if (getline(&ch_tmp, &ch_tmp_size, f) == -1)
		goto formatError;
	if (strcmp(ch_tmp, "==============================\n"))
		goto formatError;

	// - read entries
	do {
		// find beginning of next category
		do {
			tmp = fgetc(f);
		} while (tmp == '\n' && tmp != EOF);
		if (tmp == EOF)
			break;

		// decrement filepos by one
		fseek(f, -1, SEEK_CUR);
		// read cat-name
		if (getline(&ch_tmp, &ch_tmp_size, f) == -1)
			goto formatError;
		// check cat-name format
		if (memcmp(ch_tmp, "== Category: ", 13) != 0)
			goto formatError;
		if (memcmp(ch_tmp + (strlen(ch_tmp) - 1 - 3), " ==", 3) != 0)
			goto formatError;
		// copy cat-name
		curCat.assign(ch_tmp + 13, strlen(ch_tmp) - 1 - 16);

		do {
			// find beginning of next entry
			do {
				tmp = fgetc(f);
			} while (tmp == '\n' && tmp != EOF && tmp != '=');
			if (tmp == EOF)
				break;
			if (tmp == '=') {
				fseek(f, -1, SEEK_CUR);
				break;
			}
			// decrement filepos by one
			fseek(f, -1, SEEK_CUR);
			// read desc-line
			if (getline(&ch_tmp, &ch_tmp_size, f) == -1)
				goto formatError;
			// check desc-line format
			if (memcmp(ch_tmp, "-- ", 3) != 0)
				goto formatError;
			if (memcmp(ch_tmp + (strlen(ch_tmp) - 1 - 3), " --", 3) != 0)
				goto formatError;
			// add desc-line
			currItem.desc.assign(ch_tmp + 3, strlen(ch_tmp) - 1 - 6);

			// read username-line
			if ((ret = getline(&ch_tmp, &ch_tmp_size, f)) == -1)
				goto formatError;
			if (!textExtractEntry_PwM(ch_tmp, ret, &currItem.name))
				goto formatError;

			// read pw-line
			if ((ret = getline(&ch_tmp, &ch_tmp_size, f)) == -1)
				goto formatError;
			if (!textExtractEntry_PwM(ch_tmp, ret, &currItem.pw))
				goto formatError;

			// read comment-line
			if ((ret = getline(&ch_tmp, &ch_tmp_size, f)) == -1)
				goto formatError;
			if (!textExtractEntry_PwM(ch_tmp, ret, &currItem.comment))
				goto formatError;

			// read URL-line
			if ((ret = getline(&ch_tmp, &ch_tmp_size, f)) == -1)
				goto formatError;
			if (!textExtractEntry_PwM(ch_tmp, ret, &currItem.url))
				goto formatError;

			// read launcher-line
			if ((ret = getline(&ch_tmp, &ch_tmp_size, f)) == -1)
				goto formatError;
			if (!textExtractEntry_PwM(ch_tmp, ret, &currItem.launcher))
				goto formatError;

			currItem.lockStat = true;
			currItem.listViewPos = -1;
			addEntry(curCat.c_str(), &currItem, true);
			++entriesRead;
		} while (1);
	} while (1);
	if (!entriesRead)
		goto formatError;

	free(ch_tmp);
	fclose(f);
	flagDirty();
	return e_success;

      formatError:
	free(ch_tmp);
	fclose(f);
	return e_fileFormat;
}

bool PwMDoc::textExtractEntry_PwM(const char *in, ssize_t in_size, string *out)
{
	PWM_ASSERT(in && out);
	ssize_t i = 0, len = in_size - 1;
	while (i < len) {
		if (in[i] == ':')
			break;
		++i;
	}
	i += 2;
	*out = "";
	out->append(in + i, in_size - i - 1);
	return true;
}

PwMerror PwMDoc::exportToGpasman(const QString *file)
{
	PWM_ASSERT(file);
	GpasmanFile gp;
	int ret;

	if (!unlockAll_tempoary())
		return e_lock;

	QString gpmPassword;
	while (1) {
		gpmPassword = requestNewMpw(0);
		if (gpmPassword == "") {
			unlockAll_tempoary(true);
			return e_noPw;
		}
		if (gpmPassword.length() < 4) {
			gpmPwLenErrMsgBox();
		} else {
			break;
		}
	}

	ret = gp.save_init(file->latin1(), gpmPassword.latin1());
	if (ret != 1) {
		unlockAll_tempoary(true);
		return e_accessFile;
	}

	char *entry[4];
	unsigned int numCat = numCategories(), i;
	unsigned int numEntr, j;
	int descLen, nameLen, pwLen, commentLen;
	for (i = 0; i < numCat; ++i) {
		numEntr = numEntries(i);
		for (j = 0; j < numEntr; ++j) {
			descLen = dta[i].d[j].desc.length();
			nameLen = dta[i].d[j].name.length();
			pwLen = dta[i].d[j].pw.length();
			commentLen = dta[i].d[j].comment.length();
			entry[0] = new char[descLen + 1];
			entry[1] = new char[nameLen + 1];
			entry[2] = new char[pwLen + 1];
			entry[3] = new char[commentLen + 1];
			strcpy(entry[0], descLen == 0 ? " " : dta[i].d[j].desc.c_str());
			strcpy(entry[1], nameLen == 0 ? " " : dta[i].d[j].name.c_str());
			strcpy(entry[2], pwLen == 0 ? " " : dta[i].d[j].pw.c_str());
			strcpy(entry[3], commentLen == 0 ? " " : dta[i].d[j].comment.c_str());
			entry[0][descLen == 0 ? descLen + 1 : descLen] = '\0';
			entry[1][nameLen == 0 ? nameLen + 1 : nameLen] = '\0';
			entry[2][pwLen == 0 ? pwLen + 1 : pwLen] = '\0';
			entry[3][commentLen == 0 ? commentLen + 1 : commentLen] = '\0';

			ret = gp.save_entry(entry);
			if (ret == -1){
				delete [] entry[0];
				delete [] entry[1];
				delete [] entry[2];
				delete [] entry[3];
				gp.save_finalize();
				unlockAll_tempoary(true);
				return e_writeFile;
			}

			delete [] entry[0];
			delete [] entry[1];
			delete [] entry[2];
			delete [] entry[3];
		}
	}
	unlockAll_tempoary(true);
	if (gp.save_finalize() == -1)
		return e_writeFile;

	return e_success;
}

PwMerror PwMDoc::importFromGpasman(const QString *file)
{
	PWM_ASSERT(file);
	QString pw = requestMpw(false);
	if (pw == "")
		return e_noPw;
	GpasmanFile gp;
	int ret, i;
	PwMerror ret2;
	char *entry[4];
	PwMDataItem tmpData;
	ret = gp.load_init(file->latin1(), pw.latin1());
	if (ret != 1)
		return e_accessFile;

	do {
		ret = gp.load_entry(entry);
		if(ret != 1)
			break;
		tmpData.desc = entry[0];
		tmpData.name = entry[1];
		tmpData.pw = entry[2];
		tmpData.comment = entry[3];
		tmpData.lockStat = true;
		tmpData.listViewPos = -1;
		ret2 = addEntry(DEFAULT_CATEGORY, &tmpData, true);
		for (i = 0; i < 4; ++i)
         		free(entry[i]);
		if (ret2 == e_maxAllowedEntr) {
			gp.load_finalize();
			return e_maxAllowedEntr;
		}
	} while (1);
	gp.load_finalize();
	if (isDocEmpty())
		return e_wrongPw; // we assume this.

	flagDirty();
	return e_success;
}

void PwMDoc::ensureLvp()
{
	if (isDocEmpty())
		return;

	vector< vector<PwMDataItem>::iterator > undefined;
	vector< vector<PwMDataItem>::iterator >::iterator undefBegin,
							  undefEnd,
							  undefI;
	vector<PwMCategoryItem>::iterator catBegin = dta.begin(),
					  catEnd = dta.end(),
					  catI = catBegin;
	vector<PwMDataItem>::iterator entrBegin, entrEnd, entrI;
	int lvpTop, tmpLvp;

	while (catI != catEnd) {
		lvpTop = -1;
		undefined.clear();

		entrBegin = catI->d.begin();
		entrEnd = catI->d.end();
		entrI = entrBegin;

		while (entrI != entrEnd) {
			tmpLvp = entrI->listViewPos;
			if (tmpLvp == -1)
				undefined.push_back(entrI);
			else if (tmpLvp > lvpTop)
				lvpTop = tmpLvp;
			++entrI;
		}
		undefBegin = undefined.begin();
		undefEnd = undefined.end();
		undefI = undefBegin;
		while (undefI != undefEnd) {
			(*undefI)->listViewPos = ++lvpTop;
			++undefI;
		}
		++catI;
	}
}

QString PwMDoc::getTitle()
{
	/* NOTE: We have to ensure, that the returned title
	 *       is unique and not reused somewhere else while
	 *       this document is valid (open).
	 */
	QString title(getFilename());
	if (title.isEmpty()) {
		if (unnamedNum == 0) {
			unnamedNum = PwMDocList::getNewUnnamedNumber();
			PWM_ASSERT(unnamedNum != 0);
		}
		title = DEFAULT_TITLE;
		title += " ";
		title += tostr(unnamedNum).c_str();
	}
	return title;
}

#include "pwmdoc.moc"
