/***************************************************************************
 *                                                                         *
 *   copyright (C) 2003, 2004 by Michael Buesch                            *
 *   email: mbuesch@freenet.de                                             *
 *                                                                         *
 *   Original implementation of the tray-tree                              *
 *   (c) by Matt Scifo <mscifo@o1.com>                                     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License version 2        *
 *   as published by the Free Software Foundation.                         *
 *                                                                         *
 ***************************************************************************/

#include <klocale.h>

#include <qapplication.h>

#include "pwmtray.h"
#include "pwmexception.h"
#include "globalstuff.h"
#include "pwm.h"


ActiveTreeItem::ActiveTreeItem(const QString &_text, const QFont &_font,
			       ActiveItemExecTask _task, PwMDoc * _doc,
			       int _category, int _entry)
 : CustomMenuItemTempl()
 , text(_text)
 , font(_font)
 , task(_task)
 , doc(_doc)
 , category(_category)
 , entry(_entry)
{
}

ActiveTreeItem::~ActiveTreeItem()
{
}

void ActiveTreeItem::execIt()
{
	unsigned int entr = static_cast<unsigned int>(entry);
	unsigned int cat = static_cast<unsigned int>(category);

	switch (task) {
		case pwToClipboard: {
			PwMDataItem d;
			bool wasLocked = doc->isLocked(cat, entr);
			doc->getEntry(cat, entr, &d, true);
			if (wasLocked)
				doc->lockAt(cat, entr, true, false);
			PwM::copyToClipboard(d.pw.c_str());
			break;
		} case nameToClipboard: {
			PwMDataItem d;
			doc->getEntry(cat, entr, &d);
			PwM::copyToClipboard(d.name.c_str());
			break;
		} case descToClipboard: {
			PwMDataItem d;
			doc->getEntry(cat, entr, &d);
			PwM::copyToClipboard(d.desc.c_str());
			break;
		} case urlToClipboard: {
			PwMDataItem d;
			doc->getEntry(cat, entr, &d);
			PwM::copyToClipboard(d.url.c_str());
			break;
		} case launcherToClipboard: {
			PwMDataItem d;
			doc->getEntry(cat, entr, &d);
			PwM::copyToClipboard(d.launcher.c_str());
			break;
		} case commentToClipboard: {
			PwMDataItem d;
			doc->getEntry(cat, entr, &d);
			PwM::copyToClipboard(d.comment.c_str());
			break;
		} case execLauncher: {
			doc->execLauncher(cat, entr);
			break;
		} case goToURL: {
			doc->goToURL(cat, entr);
			break;
		} default: {
			BUG();
		}
	}
}



PwMTray::PwMTray(QWidget * parent, const char *name)
 : KSystemTray(parent, name)
{
	buildMain();
}

PwMTray::~PwMTray()
{
}

void PwMTray::buildMain()
{
	KPopupMenu *ctxMenu = contextMenu();
	QObject *parentWnd = parent();

	ctxMenu->insertSeparator();
	ctxMenu->insertItem(i18n("&Lock"), parentWnd,
			    SLOT(lockAllDoc_slot()));
	ctxMenu->insertItem(i18n("&Deep-Lock"), parentWnd,
			    SLOT(deepLockAllDoc_slot()));
	ctxMenu->insertSeparator();
	ctxMenu->insertItem(i18n("&Remove from tray"), parentWnd,
			    SLOT(trayUndockButton_slot()));
}

void PwMTray::insertActiveTreeItem(KPopupMenu *insertIn, const QString &text,
				   ActiveItemExecTask task, PwMDoc *doc,
				   int docCategory, int docEntry,
				   vector<int> *activeItemsList)
{
	ActiveTreeItem *activeItem;
	int id;

	activeItem = new ActiveTreeItem(text, insertIn->font(), task,
					doc, docCategory, docEntry);
	id = insertIn->insertItem(activeItem);
	insertIn->connectItem(id, activeItem, SLOT(execIt()));
	activeItemsList->push_back(id);
}

void PwMTray::rebuildTree(KPopupMenu *popup, PwMDoc *doc,
			  vector<int> *activeItems)
{
	PWM_ASSERT(doc);
	PWM_ASSERT(popup);
	activeItems->clear();
	popup->clear();

	vector<string> catList;
	vector<string> entrList;
	doc->getCategoryList(&catList);
	KPopupMenu *newCatMenu;
	KPopupMenu *newEntrMenu;
	int i, size = catList.size();
	int j, entries;
	for (i = 0; i < size; ++i) {
		newCatMenu = new KPopupMenu(popup);
		popup->insertItem(catList[i].c_str(), newCatMenu);
		doc->getEntryList(i, &entrList);

		entries = entrList.size();
		for (j = 0; j < entries; ++j) {
			newEntrMenu = new KPopupMenu(newCatMenu);
			newCatMenu->insertItem(entrList[j].c_str(), newEntrMenu);

			insertActiveTreeItem(newEntrMenu,
					     i18n("copy password to clipboard"),
					     pwToClipboard, doc, i, j,
					     activeItems);

			insertActiveTreeItem(newEntrMenu,
					     i18n("copy username to clipboard"),
					     nameToClipboard, doc, i, j,
					     activeItems);

			insertActiveTreeItem(newEntrMenu,
					     i18n("copy description to clipboard"),
					     descToClipboard, doc, i, j,
					     activeItems);

			insertActiveTreeItem(newEntrMenu,
					     i18n("copy url to clipboard"),
					     urlToClipboard, doc, i, j,
					     activeItems);

			insertActiveTreeItem(newEntrMenu,
					     i18n("copy launcher to clipboard"),
					     launcherToClipboard, doc, i, j,
					     activeItems);

			insertActiveTreeItem(newEntrMenu,
					     i18n("copy comment to clipboard"),
					     commentToClipboard, doc, i, j,
					     activeItems);

			newEntrMenu->insertSeparator();

			insertActiveTreeItem(newEntrMenu,
					     i18n("Execute \"Launcher\""),
					     execLauncher, doc, i, j,
					     activeItems);

			insertActiveTreeItem(newEntrMenu,
					     i18n("Go to \"URL\""),
					     goToURL, doc, i, j,
					     activeItems);
		}
	}
}

void PwMTray::updateTree(PwMDoc *document)
{
	PWM_ASSERT(document);
	KPopupMenu *treeEntry;
	int treeItemNum = -1;
	int id = findTreeEntry(document, &treeEntry, &treeItemNum);
	if (id == -1) {
		// tree-entry doesn't exist, yet.
		id = insertTreeEntry(document, &treeEntry);
	}
	if (treeItemNum != -1) {
		/* delete all *old* active items */
		vector<int>::iterator i;
		vector<int> *oldItems = &tree[treeItemNum].activeItems;
		KPopupMenu *ctxMenu = contextMenu();
		i = oldItems->begin();
		while (i != oldItems->end()) {
			ctxMenu->removeItem(*i);
			++i;
		}
		oldItems->clear();
		ctxMenu->changeItem(id, document->getTitle());
	}

	treeItem newTreeItem;
	if (document->isDeepLocked()) {
		treeEntry->clear();
		treeEntry->insertItem(IS_DEEPLOCKED_SHORTMSG, document,
				      SLOT(_deepUnlock()));
	} else {
		rebuildTree(treeEntry, document, &newTreeItem.activeItems);
	}
	newTreeItem.menuId = id;
	newTreeItem.doc = document;
	newTreeItem.menu = treeEntry;
	if (treeItemNum == -1) {
		tree.push_back(newTreeItem);
	} else {
		tree[treeItemNum] = newTreeItem;
	}
}

void PwMTray::closeTreeEntry(PwMDoc *document)
{
	KPopupMenu *menu;
	int treeItem;
	int id = findTreeEntry(document, &menu, &treeItem);
	if (id != -1) {
		removeTreeEntry(id, treeItem, &menu);
	}
}

int PwMTray::findTreeEntry(PwMDoc *doc, KPopupMenu **menu, int *treeItem)
{
	PWM_ASSERT(doc);
	int i, count = tree.size();
	for (i = 0; i < count; ++i) {
		if (tree[i].doc == doc) {
			if (menu)
				*menu = tree[i].menu;
			if (treeItem)
				*treeItem = i;
			return tree[i].menuId;
		}
	}
	if (menu)
		*menu = 0;
	if (treeItem)
		*treeItem = -1;
	return -1;
}

int PwMTray::insertTreeEntry(PwMDoc *doc, KPopupMenu **popup)
{
	PWM_ASSERT(doc);
	PWM_ASSERT(popup);
	KPopupMenu *ctxMenu = contextMenu();
	*popup = new KPopupMenu(contextMenu());
	return ctxMenu->insertItem(doc->getTitle(), *popup, -1, 1);
}

void PwMTray::removeTreeEntry(int menuId, int treeItem, KPopupMenu **menu)
{
	PWM_ASSERT(menu);
	KPopupMenu *ctxMenu = contextMenu();
	ctxMenu->removeItem(menuId);
	delete_and_null(*menu);
	tree.erase(tree.begin() + treeItem);
}

#include "pwmtray.moc"
