/***************************************************************************
 *                                                                         *
 *   copyright (C) 2003, 2004 by Michael Buesch                            *
 *   email: mbuesch@freenet.de                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License version 2        *
 *   as published by the Free Software Foundation.                         *
 *                                                                         *
 ***************************************************************************/

#include "pwmview.h"
#include "pwmexception.h"
#include "globalstuff.h"
#include "pwm.h"
#include "rencatwnd.h"
#include "configuration.h"
#include "commentbox.h"

#include <kmessagebox.h>
#include <klocale.h>

#include <qlineedit.h>
#include <qpoint.h>
#include <qapplication.h>

#define COLUMN_DESC		0
#define COLUMN_NAME		1
#define COLUMN_PW		2
#define COLUMN_URL		3
#define COLUMN_LAUNCHER		4

#define INITIAL_SIZE			(QSize(700, 400))


PwMView::PwMView(PwM *_mainClass,
		 QWidget *parent, PwMDoc *_doc,
		 const char *name)
 : PwMViewStyle(parent, name)
{
	PWM_ASSERT(_mainClass);
	PWM_ASSERT(parent);
	PWM_ASSERT(_doc);
	setView(this);
	doc = _doc;
	connect(doc, SIGNAL(docClosed(PwMDoc *)),
		this, SLOT(docClosed(PwMDoc *)));
	doc->setListViewPointer(this);
	mainClass = _mainClass;

	resize(INITIAL_SIZE);
	initStyle(conf()->confWndMainViewStyle());
	initCtxMenu();
	doc->setCurrentView(this);
	connect(doc, SIGNAL(dataChanged(PwMDoc *)), this, SLOT(updateView()));
}

PwMView::~PwMView()
{
	disconnect(doc, SIGNAL(docClosed(PwMDoc *)),
		   this, SLOT(docClosed(PwMDoc *)));
	delete_ifnot_null(doc);
	emit closed(this);
}

void PwMView::initCtxMenu()
{
	ctxMenu = new QPopupMenu(this);
	ctxMenu->insertItem(i18n("&Add password"), mainClass, SLOT(addPwd_slot()));
	ctxMenu->insertSeparator();
	ctxMenu->insertItem(i18n("&Edit"), mainClass, SLOT(editPwd_slot()));
	ctxMenu->insertItem(i18n("&Delete"), mainClass, SLOT(deletePwd_slot()));
	ctxMenu->insertSeparator();
	ctxMenu->insertItem(i18n("copy password to clipboard"),
			    this, SLOT(copyPwToClip()));
	ctxMenu->insertItem(i18n("copy username to clipboard"),
			    this, SLOT(copyNameToClip()));
	ctxMenu->insertItem(i18n("copy description to clipboard"),
			    this, SLOT(copyDescToClip()));
	ctxMenu->insertItem(i18n("copy url to clipboard"),
			    this, SLOT(copyUrlToClip()));
	ctxMenu->insertItem(i18n("copy launcher to clipboard"),
			    this, SLOT(copyLauncherToClip()));
	ctxMenu->insertItem(i18n("copy comment to clipboard"),
			    this, SLOT(copyCommentToClip()));
	ctxMenu->insertSeparator();
	ctxMenu->insertItem(i18n("Execute \"Launcher\""), mainClass,
			    SLOT(execLauncher_slot()));
	ctxMenu->insertItem(i18n("Go to \"URL\""), mainClass,
			    SLOT(goToURL_slot()));
}

void PwMView::docClosed(PwMDoc *_doc)
{
	if (_doc != doc)
		BUG();
	close();
}

void PwMView::closeEvent(QCloseEvent *e)
{
	if (doc->isDirty()) {
		int ret =
		    KMessageBox::questionYesNoCancel(this,
						     i18n("The list \"") +
						     caption() +
						     i18n
						     ("\" has been modified.\n"
						      "Do you want to save it?"),
						     i18n("save?"));
		if (ret == KMessageBox::Yes){	// save to disk
			mainClass->save_slot();
//			return;
		}
		else if (ret == KMessageBox::Cancel)
			return;
	}
	e->accept();
	delete this;
}

void PwMView::resizeEvent(QResizeEvent *)
{
	resizeView(size());
}

void PwMView::refreshCommentTextEdit(QListViewItem *curItem)
{
	PWM_ASSERT(commentBox);
	string comment;
	document()->getCommentByLvp(getCurrentCategory(),
				    lv->childCount() - lv->itemIndex(curItem) - 1,
				    &comment);
	commentBox->setContent(comment.c_str());
	lv->ensureItemVisible(lv->selectedItem());
}

bool PwMView::getCurEntryIndex(unsigned int *index)
{
	QListViewItem *current = lv->currentItem();
	if (!current)
		return false;

	return getDocEntryIndex(index, current);
}

bool PwMView::getDocEntryIndex(unsigned int *index,
			       const QListViewItem *item)
{
	vector<unsigned int> foundPositions;
	PwMDataItem curItem;
	curItem.desc = item->text(COLUMN_DESC).latin1();
	curItem.name = item->text(COLUMN_NAME).latin1();
	document()->getCommentByLvp(getCurrentCategory(),
				    lv->childCount() - lv->itemIndex(item) - 1,
				    &curItem.comment);
	curItem.url = item->text(COLUMN_URL).latin1();
	curItem.launcher = item->text(COLUMN_LAUNCHER).latin1();
	document()->findEntry(getCurrentCategory(), curItem, SEARCH_IN_DESC |
			      SEARCH_IN_NAME | SEARCH_IN_COMMENT | SEARCH_IN_URL |
			      SEARCH_IN_LAUNCHER,
			      &foundPositions, true);
	if (foundPositions.size()) {
		*index = foundPositions[0];
		return true;
	}

	return false;
}

void PwMView::handleToggle(QListViewItem *item)
{
	PWM_ASSERT(doc);
	if (!item)
		return;
	QCheckListItem *clItem = (QCheckListItem *)item;
	QString curCat(getCurrentCategory());

	// find document position of this entry.
	unsigned int curEntryDocIndex;
	if (!getDocEntryIndex(&curEntryDocIndex, item))
		return;

	// hack to refresh the comment, if only one item is present
	if (lv->childCount() == 1)
		refreshCommentTextEdit(lv->currentItem());

	if (doc->isLocked(curCat, curEntryDocIndex) != clItem->isOn())
		return;		// this is just a click somewhere on the entry
	if (doc->isDeepLocked()) {
		PwMerror ret;
		ret = doc->deepLock(false);
		if (ret != e_success)
			clItem->setOn(false);
		return;
	}
	doc->lockAt(curCat, curEntryDocIndex, !clItem->isOn());
}

void PwMView::handleRightClick(QListViewItem *item, const QPoint &point, int)
{
	if (!item)
		return;
	ctxMenu->move(point);
	/* don't use ctxMenu->exec() here, as it generates race conditions
	 * with the card interface code. Believe it or not. :)
	 */
	ctxMenu->show();
}

void PwMView::updateCategories()
{
	QString oldSel(getCurrentCategory());
	delAllCategories();
	QStringList catList;
	document()->getCategoryList(&catList);
	catList.sort();
	QStringList::iterator i = catList.begin(),
			      end = catList.end();
	while (i != end) {
		addCategory(*i);
		++i;
	}
	selectCategory(oldSel);
}

void PwMView::shiftToView()
{
	int cX = lv->contentsX();
	int cY = lv->contentsY();
	commentBox->clear();
	setCaption(doc->getTitle());

	unsigned int catDocIndex;
	if (unlikely(
	    !(document()->findCategory(getCurrentCategory(),
	    			       &catDocIndex)))) {
		BUG();
	}

	// ensure all listViewPos are set
	doc->ensureLvp();

	// clear all tmp-data vectors
	unsigned int i, entries = doc->numEntries(catDocIndex);
	vector<PwMDataItem> tmpSorted;
	PwMDataItem currItem;
	currItem.clear();
	tmpSorted.insert(tmpSorted.begin(), entries, currItem);

	// Sort items and store them in tempoary tmpSorted.
	for (i = 0; i < entries; ++i) {
		doc->getEntry(catDocIndex, i, &currItem);
		tmpSorted[currItem.listViewPos] = currItem;
	}

	// shift tempoary data to ListView.
	tmpDisableSort();
	lv->clear();
	QCheckListItem *newItem;
	vector<PwMDataItem>::iterator it = tmpSorted.begin(),
				      end = tmpSorted.end();
	while (it != end) {
		newItem = new QCheckListItem(lv, "", QCheckListItem::CheckBox);
		newItem->setText(COLUMN_DESC, (*it).desc.c_str());
		newItem->setText(COLUMN_NAME, (*it).name.c_str());
		newItem->setText(COLUMN_PW, (*it).pw.c_str());
		newItem->setText(COLUMN_URL, (*it).url.c_str());
		newItem->setText(COLUMN_LAUNCHER, (*it).launcher.c_str());
		newItem->setOn(!((*it).lockStat));
		lv->insertItem(newItem);
		++it;
	}
	tmpReEnableSort();

	if (cY || cX)
		lv->setContentsPos(cX, cY);
}

void PwMView::reorgLp()
{
	if (!lv->childCount())
		return;
	PWM_ASSERT(doc);
	PWM_ASSERT(!doc->isDocEmpty());
	QListViewItem *currItem;
	vector<unsigned int> foundPos;
	const int searchIn = SEARCH_IN_DESC | SEARCH_IN_NAME |
			     SEARCH_IN_URL | SEARCH_IN_LAUNCHER;
	QString curCat(getCurrentCategory());
	PwMDataItem findThis;
	unsigned int i, cnt = lv->childCount();
	for (i = 0; i < cnt; ++i) {
		currItem = lv->itemAtIndex(i);
		findThis.desc = currItem->text(COLUMN_DESC).latin1();
		findThis.name = currItem->text(COLUMN_NAME).latin1();
		findThis.url = currItem->text(COLUMN_URL).latin1();
		findThis.launcher = currItem->text(COLUMN_LAUNCHER).latin1();
		doc->findEntry(curCat, findThis, searchIn,
			       &foundPos, true);
		if (unlikely(!foundPos.size()))
			BUG();
		doc->setListViewPos(curCat, foundPos[0], cnt - i - 1);
	}
}

void PwMView::selAt(int index)
{
	QListViewItem *item = lv->itemAtIndex(index);
	if (!item)
		return;
	lv->setCurrentItem(item);
	lv->ensureItemVisible(item);
}

void PwMView::renCatButton_slot()
{
	RenCatWnd wnd(this);
	if (wnd.exec() == 1) {
		QString newName(wnd.getNewName());
		if (newName == "")
			return;
		document()->renameCategory(getCurrentCategory(),
					   newName);
	}
}

void PwMView::delCatButton_slot()
{
	if (numCategories() <= 1) {
		mainClass->showStatMsg(i18n("Can't remove the last category."));
		return;
	}
	if (KMessageBox::questionYesNo(this,
				       i18n("Do you really want to "
					    "delete the selected "
					    "category? All password-"
					    "entries will be lost in "
					    "this category!"),
				       i18n("Delete category?"))
	    == KMessageBox::No) {
		return;
	}
	document()->delCategory(getCurrentCategory());
}

void PwMView::copyPwToClip()
{
	unsigned int curIndex = 0;
	if (!getCurEntryIndex(&curIndex))
		return;
	PwMDataItem d;
	document()->getEntry(getCurrentCategory(), curIndex, &d, true);
	PwM::copyToClipboard(d.pw.c_str());
}

void PwMView::copyNameToClip()
{
	unsigned int curIndex = 0;
	if (!getCurEntryIndex(&curIndex))
		return;
	PwMDataItem d;
	document()->getEntry(getCurrentCategory(), curIndex, &d);
	PwM::copyToClipboard(d.name.c_str());
}

void PwMView::copyDescToClip()
{
	unsigned int curIndex = 0;
	if (!getCurEntryIndex(&curIndex))
		return;
	PwMDataItem d;
	document()->getEntry(getCurrentCategory(), curIndex, &d);
	PwM::copyToClipboard(d.desc.c_str());
}

void PwMView::copyUrlToClip()
{
	unsigned int curIndex = 0;
	if (!getCurEntryIndex(&curIndex))
		return;
	PwMDataItem d;
	document()->getEntry(getCurrentCategory(), curIndex, &d);
	PwM::copyToClipboard(d.url.c_str());
}

void PwMView::copyLauncherToClip()
{
	unsigned int curIndex = 0;
	if (!getCurEntryIndex(&curIndex))
		return;
	PwMDataItem d;
	document()->getEntry(getCurrentCategory(), curIndex, &d);
	PwM::copyToClipboard(d.launcher.c_str());
}

void PwMView::copyCommentToClip()
{
	unsigned int curIndex = 0;
	if (!getCurEntryIndex(&curIndex))
		return;
	PwMDataItem d;
	document()->getEntry(getCurrentCategory(), curIndex, &d);
	PwM::copyToClipboard(d.comment.c_str());
}

#include "pwmview.moc"
