/***************************************************************************
 *                                                                         *
 *   copyright (C) 2004 by Michael Buesch                                  *
 *   email: mbuesch@freenet.de                                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License version 2        *
 *   as published by the Free Software Foundation.                         *
 *                                                                         *
 ***************************************************************************/

#include "serializer.h"
#include "pwmexception.h"


/** magic string for identification */
#define ROOT_MAGIC_NAME		"PwM-xml-dat"
/** data compatibility version number.
  * All "PwM-xml-dat" with the same COMPAT_VER
  * are 100% forward- and backward-compatible.
  * So only change this version number if compatibility
  * is broken.
  */
#define COMPAT_VER		"0x02"
/** enable/disable serializer debugging (0/1) */
#define SERIALIZER_DEBUG	0


Serializer::Serializer()
{
	defaultLockStat = true;
	domDoc = new QDomDocument;
}

Serializer::Serializer(const QCString &buffer)
{
	defaultLockStat = true;
	domDoc = new QDomDocument;
	if (!parseXml(buffer)) {
		throw PwMException(EX_PARSE);
		return;
	}
}

Serializer::~Serializer()
{
	delete_ifnot_null(domDoc);
}

void Serializer::clear()
{
	delete_ifnot_null(domDoc);
	domDoc = new QDomDocument;
}

bool Serializer::parseXml(const QCString &buffer)
{
	PWM_ASSERT(domDoc);
	if (!domDoc->setContent(buffer, true))
		return false;
	if (!checkValid())
		return false;
	return true;
}

QCString Serializer::getXml()
{
	PWM_ASSERT(domDoc);

#if defined(PWM_DEBUG) && SERIALIZER_DEBUG != 0
	QCString tmp(domDoc->toCString(8));
	printDebug("<BEGIN Serializer::getXml() dump>\n");
	cout << tmp << endl;
	printDebug("<END Serializer::getXml() dump>");
#endif // DEBUG

	QCString ret(domDoc->toCString(0));
	ret.replace('\n', "");
	return ret;
}

bool Serializer::serialize(const vector<PwMCategoryItem> &dta)
{
	PWM_ASSERT(domDoc);
	QDomElement root(genNewRoot());
	QDomElement catNode(domDoc->createElement("categories"));
	root.appendChild(catNode);
	if (!addCategories(&catNode, dta))
		return false;
	return true;
}

bool Serializer::deSerialize(vector<PwMCategoryItem> *dta)
{
	PWM_ASSERT(domDoc);
	PWM_ASSERT(dta);
	QDomElement root(domDoc->documentElement());
	QDomNode n;

	dta->clear();
	for (n = root.firstChild(); !n.isNull(); n = n.nextSibling()) {
		// find <categories> ... </categories>
		if (n.nodeName() == "categories") {
			if (!readCategories(n, dta)) {
				return false;
			}

			/* NOTE: We can stop processing here, as we
			 *       don't have more nodes in root, yet.
			 */
			return true;
		}
	}
	return true;
}

bool Serializer::readCategories(const QDomNode &n,
				vector<PwMCategoryItem> *dta)
{
	QDomNodeList nl(n.childNodes());
	QDomNode cur;
	QString name;
	unsigned int numCat = nl.count(), i;
	PwMCategoryItem curCat;
	vector<PwMDataItem> curEntr;

	if (!numCat) {
		printDebug("Serializer::readCategories(): empty");
		return false;
	}
	for (i = 0; i < numCat; ++i) {
		cur = nl.item(i);
		if (cur.nodeName().left(4) == "cat_") {
			name = cur.toElement().attribute("name");
			PWM_ASSERT(name != QString::null);
			PWM_ASSERT(name != "");
			curCat.clear();
			curCat.name = name.latin1();
			if (!readEntries(cur, &curEntr)) {
				dta->clear();
				return false;
			}
			curCat.d = curEntr;
			dta->push_back(curCat);
		} else {
			printDebug("Serializer::readCategories(): uh? not a category?");
		}
	}
	return true;
}

bool Serializer::readEntries(const QDomNode &n,
			     vector<PwMDataItem> *dta)
{
	QDomNodeList nl(n.childNodes());
	QDomNode cur;
	unsigned int numEntr = nl.count(), i;
	PwMDataItem curEntr;

	dta->clear();
	for (i = 0; i < numEntr; ++i) {
		cur = nl.item(i);
		if (cur.nodeName().left(6) == "entry_") {
			if (!extractEntry(cur, &curEntr)) {
				return false;
			}
			dta->push_back(curEntr);
		} else {
			printDebug("Serializer::readEntries(): hm? not an entry?");
		}
	}
	return true;
}

bool Serializer::extractEntry(const QDomNode &n,
			      PwMDataItem *dta)
{
	QDomNodeList nl(n.childNodes());
	QDomNode cur, cdata;
	unsigned int cnt = nl.count(), i;
	QString name, text;

	if (!cnt) {
		printDebug("Serializer::extractEntry(): empty");
		return false;
	}
	dta->clear();
	for (i = 0; i < cnt; ++i) {
		cur = nl.item(i);
		name = cur.nodeName();
		cdata = cur.firstChild();
		if (likely(cdata.isCDATASection())) {
			text = cdata.toCDATASection().data();
		} else if (cur.isElement()) {
			// This is for backward compatibility.
			text = cur.toElement().text();
		} else {
			printDebug("Serializer::extractEntry(): neither CDATA nor element.");
			return false;
		}
		if (text == " ")
			text = ""; // for backward compatibility.
		if (name == "desc") {
			dta->desc = unescapeEntryData(text).latin1();
		} else if (name == "name") {
			dta->name = unescapeEntryData(text).latin1();
		} else if (name == "pw") {
			dta->pw = unescapeEntryData(text).latin1();
		} else if (name == "comment") {
			dta->comment = unescapeEntryData(text).latin1();
		} else if (name == "url") {
			dta->url = unescapeEntryData(text).latin1();
		} else if (name == "launcher") {
			dta->launcher = unescapeEntryData(text).latin1();
		} else if (name == "listViewPos") {
			dta->listViewPos = strtol(text.latin1(), 0, 10);
		} else {
			printDebug(string("Serializer::extractEntry(): invalid: ")
				   + name.latin1());
		}
	}
	dta->lockStat = defaultLockStat;
	return true;
}

bool Serializer::checkValid()
{
	PWM_ASSERT(domDoc);
	QDomElement root(domDoc->documentElement());
	if (root.nodeName() != ROOT_MAGIC_NAME) {
		printDebug("Serializer: wrong magic");
		return false;
	}
	if (root.attribute("ver") != COMPAT_VER) {
		printDebug("Serializer: wrong version");
		return false;
	}
	return true;
}

QDomElement Serializer::genNewRoot()
{
	PWM_ASSERT(domDoc);
	QDomElement root(domDoc->createElement(ROOT_MAGIC_NAME));
	root.setAttribute("ver", COMPAT_VER);
	domDoc->appendChild(root);
	return root;
}

bool Serializer::addCategories(QDomElement *e,
			       const vector<PwMCategoryItem> &dta)
{
	unsigned int numCat = dta.size(), i;
	QString curId, curName;
	QDomElement curCat;

	for (i = 0; i < numCat; ++i) {
		curId = "cat_";
		curId += tostr(i).c_str();
		curName = dta[i].name.c_str();
		curCat = domDoc->createElement(curId);
		curCat.setAttribute("name", curName);
		if (!addEntries(&curCat, dta[i].d)) {
			return false;
		}
		e->appendChild(curCat);
	}
	return true;
}

bool Serializer::addEntries(QDomElement *e,
			    const vector<PwMDataItem> &dta)
{
	unsigned int numEntr = dta.size(), i;
	QString curId;
	QDomElement curEntr;

	for (i = 0; i < numEntr; ++i) {
		curId = "entry_";
		curId += tostr(i).c_str();
		curEntr = domDoc->createElement(curId);
		if (!writeEntry(&curEntr, dta[i])) {
			return false;
		}
		e->appendChild(curEntr);
	}
	return true;
}

bool Serializer::writeEntry(QDomElement *e,
			    const PwMDataItem &_dta)
{
	QDomCDATASection curText;
	QDomElement tag;

	// begin -- This is for compatibility with the old serializer
	PwMDataItem dta = _dta;
	if (!dta.desc.size())
		dta.desc = " ";
	if (!dta.name.size())
		dta.name = " ";
	if (!dta.pw.size())
		dta.pw = " ";
	if (!dta.comment.size())
		dta.comment = " ";
	if (!dta.url.size())
		dta.url = " ";
	if (!dta.launcher.size())
		dta.launcher = " ";
	// end -- This is for compatibility with the old serializer

	tag = domDoc->createElement("desc");
	curText = domDoc->createCDATASection(escapeEntryData(dta.desc.c_str()));
	tag.appendChild(curText);
	e->appendChild(tag);

	tag = domDoc->createElement("name");
	curText = domDoc->createCDATASection(escapeEntryData(dta.name.c_str()));
	tag.appendChild(curText);
	e->appendChild(tag);

	tag = domDoc->createElement("pw");
	curText = domDoc->createCDATASection(escapeEntryData(dta.pw.c_str()));
	tag.appendChild(curText);
	e->appendChild(tag);

	tag = domDoc->createElement("comment");
	curText = domDoc->createCDATASection(escapeEntryData(dta.comment.c_str()));
	tag.appendChild(curText);
	e->appendChild(tag);

	tag = domDoc->createElement("url");
	curText = domDoc->createCDATASection(escapeEntryData(dta.url.c_str()));
	tag.appendChild(curText);
	e->appendChild(tag);

	tag = domDoc->createElement("launcher");
	curText = domDoc->createCDATASection(escapeEntryData(dta.launcher.c_str()));
	tag.appendChild(curText);
	e->appendChild(tag);

	tag = domDoc->createElement("listViewPos");
	curText = domDoc->createCDATASection(tostr(dta.listViewPos).c_str());
	tag.appendChild(curText);
	e->appendChild(tag);

	return true;
}

QString Serializer::escapeEntryData(QString dta)
{
	dta.replace('\n', "$>--endl--<$");
	dta.replace("]]>", "||>");
	return dta;
}

QString Serializer::unescapeEntryData(QString dta)
{
	dta.replace("$>--endl--<$", "\n");
	dta.replace("||>", "]]>");
	return dta;
}
